# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""gem dataset."""

from __future__ import annotations

import csv
import json
import os
import textwrap

from etils import epath
import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
**GEM** is a benchmark environment for Natural Language Generation with a focus
on its Evaluation, both through human annotations and automated Metrics.

GEM aims to: (1) measure NLG progress across 13 datasets spanning many NLG
tasks and languages. (2) provide an in-depth analysis of data and models
presented via data statements and challenge sets. (3) develop standards for
evaluation of generated text using both automated and human metrics.

More information can be found at
[https://gem-benchmark.com](https://gem-benchmark.com).
"""

_URL = "https://gem-benchmark.com"

_CITATION = r"""@article{gehrmann2021gem,
  author    = {Sebastian Gehrmann and
               Tosin P. Adewumi and
               Karmanya Aggarwal and
               Pawan Sasanka Ammanamanchi and
               Aremu Anuoluwapo and
               Antoine Bosselut and
               Khyathi Raghavi Chandu and
               Miruna{-}Adriana Clinciu and
               Dipanjan Das and
               Kaustubh D. Dhole and
               Wanyu Du and
               Esin Durmus and
               Ondrej Dusek and
               Chris Emezue and
               Varun Gangal and
               Cristina Garbacea and
               Tatsunori Hashimoto and
               Yufang Hou and
               Yacine Jernite and
               Harsh Jhamtani and
               Yangfeng Ji and
               Shailza Jolly and
               Dhruv Kumar and
               Faisal Ladhak and
               Aman Madaan and
               Mounica Maddela and
               Khyati Mahajan and
               Saad Mahamood and
               Bodhisattwa Prasad Majumder and
               Pedro Henrique Martins and
               Angelina McMillan{-}Major and
               Simon Mille and
               Emiel van Miltenburg and
               Moin Nadeem and
               Shashi Narayan and
               Vitaly Nikolaev and
               Rubungo Andre Niyongabo and
               Salomey Osei and
               Ankur P. Parikh and
               Laura Perez{-}Beltrachini and
               Niranjan Ramesh Rao and
               Vikas Raunak and
               Juan Diego Rodriguez and
               Sashank Santhanam and
               Jo{\~{a}}o Sedoc and
               Thibault Sellam and
               Samira Shaikh and
               Anastasia Shimorina and
               Marco Antonio Sobrevilla Cabezudo and
               Hendrik Strobelt and
               Nishant Subramani and
               Wei Xu and
               Diyi Yang and
               Akhila Yerukola and
               Jiawei Zhou},
  title     = {The {GEM} Benchmark: Natural Language Generation, its Evaluation and
               Metrics},
  journal   = {CoRR},
  volume    = {abs/2102.01672},
  year      = {2021},
  url       = {https://arxiv.org/abs/2102.01672},
  archivePrefix = {arXiv},
  eprint    = {2102.01672}
}

Note that each GEM dataset has its own citation. Please see the source to see
the correct citation for each contained dataset."
"""

# Used to convert IDs into dialog act names in schema-guided dialog.
_SGD_ACTS = [
    "AFFIRM",
    "AFFIRM_INTENT",
    "CONFIRM",
    "GOODBYE",
    "INFORM",
    "INFORM_COUNT",
    "INFORM_INTENT",
    "NEGATE",
    "NEGATE_INTENT",
    "NOTIFY_FAILURE",
    "NOTIFY_SUCCESS",
    "OFFER",
    "OFFER_INTENT",
    "REQUEST",
    "REQUEST_ALTS",
    "REQ_MORE",
    "SELECT",
    "THANK_YOU",
]

# Used in Xsum to clean data.
_XSUM_REMOVE_LINES = set([
    "Share this with\n",
    "Email\n",
    "Facebook\n",
    "Messenger\n",
    "Twitter\n",
    "Pinterest\n",
    "WhatsApp\n",
    "Linkedin\n",
    "LinkedIn\n",
    "Copy this link\n",
    "These are external links and will open in a new window\n",
])

_WIKI_LINGUA_LANGS = {
    "wiki_lingua_arabic_ar": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/arabic.zip",
    "wiki_lingua_chinese_zh": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/chinese.zip",
    "wiki_lingua_czech_cs": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/czech.zip",
    "wiki_lingua_dutch_nl": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/dutch.zip",
    "wiki_lingua_english_en": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/english.zip",
    "wiki_lingua_french_fr": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/french.zip",
    "wiki_lingua_german_de": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/german.zip",
    "wiki_lingua_hindi_hi": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/hindi.zip",
    "wiki_lingua_indonesian_id": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/indonesian.zip",
    "wiki_lingua_italian_it": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/italian.zip",
    "wiki_lingua_japanese_ja": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/japanese.zip",
    "wiki_lingua_korean_ko": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/korean.zip",
    "wiki_lingua_portuguese_pt": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/portuguese.zip",
    "wiki_lingua_russian_ru": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/russian.zip",
    "wiki_lingua_spanish_es": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/spanish.zip",
    "wiki_lingua_thai_th": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/thai.zip",
    "wiki_lingua_turkish_tr": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/turkish.zip",
    "wiki_lingua_vietnamese_vi": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_wikilingua_full/vietnamese.zip",
}


class GemConfig(tfds.core.BuilderConfig):
  """BuilderConfig for GEM."""

  def __init__(self, *, features, data_urls, citation, **kwargs):
    """BuilderConfig for GEM.

    Args:
      features: `tfds.features.FeaturesDict`, specific feature dict for the
        dataset.
      data_urls: `dict`, urls to download the files from
      citation: `string`, citation for the data set
      **kwargs: keyword arguments forwarded to super.
    """
    super(GemConfig, self).__init__(**kwargs)
    self.features = features
    self.data_urls = data_urls
    self.citation = citation


class Gem(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for GEM benchmark."""

  VERSION = tfds.core.Version("1.1.0")
  RELEASE_NOTES = {
      "1.1.0": "Release of the Challenge Sets",
      "1.0.1": "Update bad links filter for MLSum",
      "1.0.0": "Initial version",
  }
  BUILDER_CONFIGS = [
      GemConfig(
          name="common_gen",
          description=textwrap.dedent(
              """\
      CommonGen is a constrained text generation task, associated with a
      benchmark dataset, to explicitly test machines for the ability of
      generative commonsense reasoning. Given a set of common concepts;
      the task is to generate a coherent sentence describing an everyday
      scenario using these concepts.
      """
          ),
          features=tfds.features.FeaturesDict({
              "gem_id": np.str_,
              "gem_parent_id": np.str_,
              "concept_set_id": np.int32,
              "concepts": tfds.features.Sequence(np.str_),
              "target": np.str_,  # single target for train.
              "references": tfds.features.Sequence(
                  np.str_
              ),  # multiple references for validation.
          }),
          data_urls={
              "data": "https://storage.googleapis.com/huggingface-nlp/datasets/common_gen/commongen_data.zip",
              "challenge_set": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_challenge_sets/common_gen.zip",
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{lin2020commongen,
            title = "CommonGen: A Constrained Text Generation Challenge for Generative Commonsense Reasoning",
            author = "Lin, Bill Yuchen  and
              Zhou, Wangchunshu  and
              Shen, Ming  and
              Zhou, Pei  and
              Bhagavatula, Chandra  and
              Choi, Yejin  and
              Ren, Xiang",
            booktitle = "Findings of the Association for Computational Linguistics: EMNLP 2020",
            month = nov,
            year = "2020",
            address = "Online",
            publisher = "Association for Computational Linguistics",
            url = "https://www.aclweb.org/anthology/2020.findings-emnlp.165",
            pages = "1823--1840",
          }"""
          ),
      ),
      GemConfig(
          name="cs_restaurants",
          description=textwrap.dedent(
              """\
          The task is generating responses in the context of a (hypothetical)
          dialogue system that provides information about restaurants.
          The input is a basic intent/dialogue act type and a list of slots
          (attributes) and their values. The output is a natural language
          sentence."""
          ),
          features=tfds.features.FeaturesDict({
              "gem_id": np.str_,
              "gem_parent_id": np.str_,
              "dialog_act": np.str_,
              "dialog_act_delexicalized": np.str_,
              "target_delexicalized": np.str_,
              "target": np.str_,  # single target for train.
              "references": tfds.features.Sequence(
                  np.str_
              ),  # multiple references for validation.
          }),
          data_urls={
              "train": "https://raw.githubusercontent.com/UFAL-DSG/cs_restaurant_dataset/master/train.json",
              "validation": "https://raw.githubusercontent.com/UFAL-DSG/cs_restaurant_dataset/master/devel.json",
              "test": "https://raw.githubusercontent.com/UFAL-DSG/cs_restaurant_dataset/master/test.json",
              "challenge_set": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_challenge_sets/cs_restaurants.zip",
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{cs_restaurants,
            address = {Tokyo, Japan},
            title = {Neural {Generation} for {Czech}: {Data} and {Baselines}},
            shorttitle = {Neural {Generation} for {Czech}},
            url = {https://www.aclweb.org/anthology/W19-8670/},
            urldate = {2019-10-18},
            booktitle = {Proceedings of the 12th {International} {Conference} on {Natural} {Language} {Generation} ({INLG} 2019)},
            author = {Dušek, Ondřej and Jurčíček, Filip},
            month = oct,
            year = {2019},
            pages = {563--574}
          }"""
          ),
      ),
      GemConfig(
          name="dart",
          description=textwrap.dedent(
              """\
            DART is a large and open-domain structured DAta Record to Text
            generation corpus with high-quality sentence annotations with each
            input being a set of entity-relation triples following a
            tree-structured ontology."""
          ),
          features=tfds.features.FeaturesDict({
              "gem_id": np.str_,
              "gem_parent_id": np.str_,
              "dart_id": np.int32,
              "tripleset": tfds.features.Sequence(np.str_),
              "subtree_was_extended": np.bool_,
              "target_sources": tfds.features.Sequence(np.str_),
              "target": np.str_,  # single target for train.
              "references": tfds.features.Sequence(
                  np.str_
              ),  # multiple references for validation.
          }),
          data_urls={
              "train": "https://raw.githubusercontent.com/Yale-LILY/dart/master/data/v1.1.1/dart-v1.1.1-full-train.json",
              "validation": "https://raw.githubusercontent.com/Yale-LILY/dart/master/data/v1.1.1/dart-v1.1.1-full-dev.json",
              "test": "https://raw.githubusercontent.com/Yale-LILY/dart/master/data/v1.1.1/dart-v1.1.1-full-test.json",
          },
          citation=textwrap.dedent(
              """\
            @article{radev2020dart,
              title=Dart: Open-domain structured data record to text generation,
              author={Radev, Dragomir and Zhang, Rui and Rau, Amrit and Sivaprasad, Abhinand and Hsieh, Chiachun and Rajani, Nazneen Fatema and Tang, Xiangru and Vyas, Aadit and Verma, Neha and Krishna, Pranav and others},
              journal={arXiv preprint arXiv:2007.02871},
              year={2020}
            }
            }"""
          ),
      ),
      GemConfig(
          name="e2e_nlg",
          description=textwrap.dedent(
              """\
            The E2E dataset is designed for a limited-domain data-to-text task
            -- generation of restaurant descriptions/recommendations based on up
            to 8 different attributes (name, area, price range etc.)"""
          ),
          features=tfds.features.FeaturesDict({
              "gem_id": np.str_,
              "gem_parent_id": np.str_,
              "meaning_representation": np.str_,
              "target": np.str_,  # single target for train.
              "references": tfds.features.Sequence(
                  np.str_
              ),  # multiple references for validation.
          }),
          data_urls={
              "train": "https://github.com/tuetschek/e2e-cleaning/raw/master/cleaned-data/train-fixed.no-ol.csv",
              "validation": "https://github.com/tuetschek/e2e-cleaning/raw/master/cleaned-data/devel-fixed.no-ol.csv",
              "test": "https://github.com/tuetschek/e2e-cleaning/raw/master/cleaned-data/test-fixed.csv",
              "challenge_set": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_challenge_sets/e2e_nlg.zip",
          },
          citation=textwrap.dedent(
              """\
            @inproceedings{e2e_cleaned,
              address = {Tokyo, Japan},
              title = {Semantic {Noise} {Matters} for {Neural} {Natural} {Language} {Generation}},
              url = {https://www.aclweb.org/anthology/W19-8652/},
              booktitle = {Proceedings of the 12th {International} {Conference} on {Natural} {Language} {Generation} ({INLG} 2019)},
              author = {Dušek, Ondřej and Howcroft, David M and Rieser, Verena},
              year = {2019},
              pages = {421--426},
            }"""
          ),
      ),
      GemConfig(
          name="mlsum_de",
          description=textwrap.dedent(
              """\
            MLSum is a large-scale multiLingual summarization dataset. It is
            buillt from online news outlets, this split focusing on German."""
          ),
          features=tfds.features.FeaturesDict({
              "gem_id": np.str_,
              "gem_parent_id": np.str_,
              "text": np.str_,
              "topic": np.str_,
              "url": np.str_,
              "title": np.str_,
              "date": np.str_,
              "target": np.str_,  # single target for train.
              "references": tfds.features.Sequence(
                  np.str_
              ),  # multiple references for validation.
          }),
          data_urls={
              "train": "https://gitlab.lip6.fr/scialom/mlsum_data/-/raw/master/MLSUM/de_train.zip",
              "validation": "https://gitlab.lip6.fr/scialom/mlsum_data/-/raw/master/MLSUM/de_val.zip",
              "test": "https://gitlab.lip6.fr/scialom/mlsum_data/-/raw/master/MLSUM/de_test.zip",
              "bad_ids": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_mlsum_bad_ids_fixed.json",
              "challenge_set": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_challenge_sets/mlsum_de.zip",
          },
          citation=textwrap.dedent(
              """\
            @inproceedings{scialom-etal-2020-mlsum,
                title = "{MLSUM}: The Multilingual Summarization Corpus",
                author = {Scialom, Thomas  and Dray, Paul-Alexis  and Lamprier, Sylvain  and Piwowarski, Benjamin  and Staiano, Jacopo},
                booktitle = {Proceedings of the 2020 Conference on Empirical Methods in Natural Language Processing (EMNLP)},
                year = {2020}
            }"""
          ),
      ),
      GemConfig(
          name="mlsum_es",
          description=textwrap.dedent(
              """\
            MLSum is a large-scale multiLingual summarization dataset. It is
            buillt from online news outlets, this split focusing on
            Spanish."""
          ),
          features=tfds.features.FeaturesDict({
              "gem_id": np.str_,
              "gem_parent_id": np.str_,
              "text": np.str_,
              "topic": np.str_,
              "url": np.str_,
              "title": np.str_,
              "date": np.str_,
              "target": np.str_,  # single target for train.
              "references": tfds.features.Sequence(
                  np.str_
              ),  # multiple references for validation.
          }),
          data_urls={
              "train": "https://gitlab.lip6.fr/scialom/mlsum_data/-/raw/master/MLSUM/es_train.zip",
              "validation": "https://gitlab.lip6.fr/scialom/mlsum_data/-/raw/master/MLSUM/es_val.zip",
              "test": "https://gitlab.lip6.fr/scialom/mlsum_data/-/raw/master/MLSUM/es_test.zip",
              "bad_ids": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_mlsum_bad_ids_fixed.json",
              "challenge_set": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_challenge_sets/mlsum_es.zip",
          },
          citation=textwrap.dedent(
              """\
            @inproceedings{scialom-etal-2020-mlsum,
                title = "{MLSUM}: The Multilingual Summarization Corpus",
                author = {Scialom, Thomas  and Dray, Paul-Alexis  and Lamprier, Sylvain  and Piwowarski, Benjamin  and Staiano, Jacopo},
                booktitle = {Proceedings of the 2020 Conference on Empirical Methods in Natural Language Processing (EMNLP)},
                year = {2020}
            }"""
          ),
      ),
      GemConfig(
          name="schema_guided_dialog",
          description=textwrap.dedent(
              """\
            The Schema-Guided Dialogue (SGD) dataset contains 18K multi-domain
            task-oriented dialogues between a human and a virtual assistant,
            which covers 17 domains ranging from banks and events to media,
            calendar, travel, and weather."""
          ),
          features=tfds.features.FeaturesDict({
              "gem_id": np.str_,
              "gem_parent_id": np.str_,
              "dialog_id": np.str_,
              "turn_id": np.int32,
              "service": np.str_,
              "prompt": np.str_,
              "context": tfds.features.Sequence(
                  np.str_
              ),  # multiple references for validation.
              "dialog_acts": tfds.features.Sequence({
                  "act": tfds.features.ClassLabel(names=_SGD_ACTS),
                  "slot": np.str_,
                  "values": tfds.features.Sequence(np.str_),
              }),
              "target": np.str_,  # single target for train.
              "references": tfds.features.Sequence(
                  np.str_
              ),  # multiple references for validation.
          }),
          data_urls={
              "data": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_sgd_context.zip",
              "challenge_set": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_challenge_sets/schema_guided_dialog.zip",
          },
          citation=textwrap.dedent(
              """\
            @article{rastogi2019towards,
              title={Towards Scalable Multi-domain Conversational Agents: The Schema-Guided Dialogue Dataset},
              author={Rastogi, Abhinav and Zang, Xiaoxue and Sunkara, Srinivas and Gupta, Raghav and Khaitan, Pranav},
              journal={arXiv preprint arXiv:1909.05855},
              year={2019}
            }"""
          ),
      ),
      GemConfig(
          name="totto",
          description=textwrap.dedent(
              """\
            ToTTo is a Table-to-Text NLG task. The task is as follows:
            Given a Wikipedia table with row names, column names and table
            cells, with a subset of cells highlighted, generate a natural
            language description for the highlighted part of the table."""
          ),
          features=tfds.features.FeaturesDict({
              "gem_id": np.str_,
              "gem_parent_id": np.str_,
              "totto_id": np.int32,
              "table_page_title": np.str_,
              "table_webpage_url": np.str_,
              "table_section_title": np.str_,
              "table_section_text": np.str_,
              "table": tfds.features.Sequence(
                  tfds.features.Sequence({
                      "column_span": np.int32,
                      "is_header": np.bool_,
                      "row_span": np.int32,
                      "value": np.str_,
                  })
              ),
              "highlighted_cells": tfds.features.Sequence(
                  tfds.features.Sequence(np.int32)
              ),
              "example_id": np.str_,
              "overlap_subset": np.str_,
              "sentence_annotations": tfds.features.Sequence({
                  "original_sentence": np.str_,
                  "sentence_after_deletion": np.str_,
                  "sentence_after_ambiguity": np.str_,
                  "final_sentence": np.str_,
              }),
              "target": np.str_,  # single target for train.
              "references": tfds.features.Sequence(
                  np.str_
              ),  # multiple references for validation.
          }),
          data_urls={
              "data": "https://storage.googleapis.com/totto/totto_data.zip",
              "challenge_set": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_challenge_sets/totto.zip",
          },
          citation=textwrap.dedent(
              """\
            @inproceedings{parikh2020totto,
              title=ToTTo: A Controlled Table-To-Text Generation Dataset,
              author={Parikh, Ankur and Wang, Xuezhi and Gehrmann, Sebastian and Faruqui, Manaal and Dhingra, Bhuwan and Yang, Diyi and Das, Dipanjan},
              booktitle={Proceedings of the 2020 Conference on Empirical Methods in Natural Language Processing (EMNLP)},
              pages={1173--1186},
              year={2020}
            }"""
          ),
      ),
      GemConfig(
          name="web_nlg_en",
          description=textwrap.dedent(
              """\
            WebNLG is a bi-lingual dataset (English, Russian) of parallel
            DBpedia triple sets and short texts that cover about 450 different
            DBpedia properties. The WebNLG data was originally created to
            promote the development of RDF verbalisers able to generate short
            text and to handle micro-planning."""
          ),
          features=tfds.features.FeaturesDict({
              "gem_id": np.str_,
              "gem_parent_id": np.str_,
              "input": tfds.features.Sequence(np.str_),
              "category": np.str_,
              "webnlg_id": np.str_,
              "target": np.str_,  # single target for train.
              "references": tfds.features.Sequence(
                  np.str_
              ),  # multiple references for validation.
          }),
          data_urls={
              "train": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_web_nlg/webnlg_en_train.json",
              "validation": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_web_nlg/webnlg_en_val.json",
              "test": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_web_nlg/webnlg_en_test.json",
              "challenge_set": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_challenge_sets/web_nlg_en.zip",
          },
          citation=textwrap.dedent(
              """\
            @inproceedings{gardent2017creating,
              author = "Gardent, Claire
                and Shimorina, Anastasia
                and Narayan, Shashi
                and Perez-Beltrachini, Laura",
              title = "Creating Training Corpora for NLG Micro-Planners",
              booktitle = "Proceedings of the 55th Annual Meeting of the Association for Computational Linguistics (Volume 1: Long Papers)",
              year = "2017",
              publisher = "Association for Computational Linguistics",
              pages = "179--188",
              location = "Vancouver, Canada",
              doi = "10.18653/v1/P17-1017",
              url = "http://www.aclweb.org/anthology/P17-1017"
            }"""
          ),
      ),
      GemConfig(
          name="web_nlg_ru",
          description=textwrap.dedent(
              """\
            WebNLG is a bi-lingual dataset (English, Russian) of parallel
            DBpedia triple sets and short texts that cover about 450 different
            DBpedia properties. The WebNLG data was originally created to
            promote the development of RDF verbalisers able to generate short
            text and to handle micro-planning."""
          ),
          features=tfds.features.FeaturesDict({
              "gem_id": np.str_,
              "gem_parent_id": np.str_,
              "input": tfds.features.Sequence(np.str_),
              "category": np.str_,
              "webnlg_id": np.str_,
              "target": np.str_,  # single target for train.
              "references": tfds.features.Sequence(
                  np.str_
              ),  # multiple references for validation.
          }),
          data_urls={
              "train": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_web_nlg/webnlg_ru_train.json",
              "validation": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_web_nlg/webnlg_ru_val.json",
              "test": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_web_nlg/webnlg_ru_test.json",
              "challenge_set": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_challenge_sets/web_nlg_ru.zip",
          },
          citation=textwrap.dedent(
              """\
            @inproceedings{gardent2017creating,
              author = "Gardent, Claire
                and Shimorina, Anastasia
                and Narayan, Shashi
                and Perez-Beltrachini, Laura",
              title = "Creating Training Corpora for NLG Micro-Planners",
              booktitle = "Proceedings of the 55th Annual Meeting of the Association for Computational Linguistics (Volume 1: Long Papers)",
              year = "2017",
              publisher = "Association for Computational Linguistics",
              pages = "179--188",
              location = "Vancouver, Canada",
              doi = "10.18653/v1/P17-1017",
              url = "http://www.aclweb.org/anthology/P17-1017"
            }"""
          ),
      ),
      GemConfig(
          name="wiki_auto_asset_turk",
          description=textwrap.dedent(
              """\
            WikiAuto provides a set of aligned sentences from English Wikipedia
            and Simple English Wikipedia as a resource to train sentence
            simplification systems. ASSET and TURK are high-quality
            simplification datasets used for testing."""
          ),
          features=tfds.features.FeaturesDict({
              "gem_id": np.str_,
              "gem_parent_id": np.str_,
              "source": np.str_,
              "target": np.str_,  # single target for train.
              "references": tfds.features.Sequence(
                  np.str_
              ),  # multiple references for validation.
          }),
          data_urls={
              "train": "https://github.com/chaojiang06/wiki-auto/raw/master/wiki-auto/GEM2021/full_with_split/train.tsv",
              "validation": "https://github.com/chaojiang06/wiki-auto/raw/master/wiki-auto/GEM2021/full_with_split/valid.tsv",
              "test_turk": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_turk_detokenized.json",
              "challenge_set": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_challenge_sets/wiki_auto_asset_turk_train_valid.zip",
              "test_asset_0": "https://github.com/facebookresearch/asset/raw/main/dataset/asset.test.simp.0",
              "test_asset_1": "https://github.com/facebookresearch/asset/raw/main/dataset/asset.test.simp.1",
              "test_asset_2": "https://github.com/facebookresearch/asset/raw/main/dataset/asset.test.simp.2",
              "test_asset_3": "https://github.com/facebookresearch/asset/raw/main/dataset/asset.test.simp.3",
              "test_asset_4": "https://github.com/facebookresearch/asset/raw/main/dataset/asset.test.simp.4",
              "test_asset_5": "https://github.com/facebookresearch/asset/raw/main/dataset/asset.test.simp.5",
              "test_asset_6": "https://github.com/facebookresearch/asset/raw/main/dataset/asset.test.simp.6",
              "test_asset_7": "https://github.com/facebookresearch/asset/raw/main/dataset/asset.test.simp.7",
              "test_asset_8": "https://github.com/facebookresearch/asset/raw/main/dataset/asset.test.simp.8",
              "test_asset_9": "https://github.com/facebookresearch/asset/raw/main/dataset/asset.test.simp.9",
          },
          citation=textwrap.dedent(
              """\
            @inproceedings{jiang-etal-2020-neural,
              title = "Neural {CRF} Model for Sentence Alignment in Text Simplification",
              author = "Jiang, Chao  and
                Maddela, Mounica  and
                Lan, Wuwei  and
                Zhong, Yang  and
                Xu, Wei",
              booktitle = "Proceedings of the 58th Annual Meeting of the Association for Computational Linguistics",
              month = jul,
              year = "2020",
              address = "Online",
              publisher = "Association for Computational Linguistics",
              url = "https://www.aclweb.org/anthology/2020.acl-main.709",
              doi = "10.18653/v1/2020.acl-main.709",
              pages = "7943--7960",
          }"""
          ),
      ),
      GemConfig(
          name="xsum",
          description=textwrap.dedent(
              """\
            The dataset is for the task of abstractive summarization in its
            extreme form, its about summarizing a document in a single
            sentence."""
          ),
          features=tfds.features.FeaturesDict({
              "gem_id": np.str_,
              "gem_parent_id": np.str_,
              "xsum_id": np.str_,
              "document": np.str_,
              "target": np.str_,  # single target for train.
              "references": tfds.features.Sequence(
                  np.str_
              ),  # multiple references for validation.
          }),
          data_urls={
              "data": "http://bollin.inf.ed.ac.uk/public/direct/XSUM-EMNLP18-Summary-Data-Original.tar.gz",
              "splits": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_xsum_confidence_0.8.json",
              "challenge_set": "https://storage.googleapis.com/huggingface-nlp/datasets/gem/gem_challenge_sets/xsum.zip",
          },
          citation=textwrap.dedent(
              """\
            @inproceedings{Narayan2018dont,
              author = "Shashi Narayan and Shay B. Cohen and Mirella Lapata",
              title = "Don't Give Me the Details, Just the Summary! {T}opic-Aware Convolutional Neural Networks for Extreme Summarization",
              booktitle = "Proceedings of the 2018 Conference on Empirical Methods in Natural Language Processing ",
              year = "2018",
              address = "Brussels, Belgium",
            }"""
          ),
      ),
  ]
  # Add all the WikiLingua versions.
  for version_name, data_url in _WIKI_LINGUA_LANGS.items():
    ln = version_name.split("_")[-1]
    BUILDER_CONFIGS.append(
        GemConfig(
            name=version_name,
            description=textwrap.dedent(
                """\
                Wikilingua is a large-scale, multilingual dataset for the evaluation
                of cross-lingual abstractive summarization systems.."""
            ),
            features=tfds.features.FeaturesDict({
                "gem_id": np.str_,
                "gem_parent_id": np.str_,
                "source": np.str_,
                "target": np.str_,  # single target for train.
                "source_aligned": tfds.features.Translation(
                    languages=[ln, "en"]
                ),  # parallel in English.
                "target_aligned": tfds.features.Translation(
                    languages=[ln, "en"]
                ),
                "references": tfds.features.Sequence(
                    np.str_
                ),  # multiple references for validation.
            }),
            data_urls={"data": data_url},
            citation=textwrap.dedent(
                """\
                @inproceedings{ladhak-wiki-2020,
                title=WikiLingua: A New Benchmark Dataset for Multilingual Abstractive Summarization,
                author={Faisal Ladhak, Esin Durmus, Claire Cardie and Kathleen McKeown},
                booktitle={Findings of EMNLP, 2020},
                year={2020}
                }"""
            ),
        )
    )

  def _info(self) -> tfds.core.DatasetInfo:
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict(self.builder_config.features),
        homepage=_URL,
        citation=self.builder_config.citation + "\n" + _CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    files = dl_manager.download_and_extract(self.builder_config.data_urls)
    if self.builder_config.name == "common_gen":
      challenge_sets = [
          ("challenge_train_sample", "train_common_gen_RandomSample500.json"),
          (
              "challenge_validation_sample",
              "validation_common_gen_RandomSample500.json",
          ),
          (
              "challenge_test_scramble",
              "test_common_gen_ScrambleInputStructure500.json",
          ),
      ]

      challenge_splits = []
      for challenge_split, filename in challenge_sets:
        challenge_splits.append(
            tfds.core.SplitGenerator(
                name=challenge_split,
                gen_kwargs={
                    "filepath": os.path.join(
                        files["challenge_set"],
                        self.builder_config.name,
                        filename,
                    ),
                    "set_name": challenge_split,
                },
            )
        )

      return [
          tfds.core.SplitGenerator(
              name=tfds.Split.TRAIN,
              gen_kwargs={
                  "filepath": os.path.join(
                      files["data"], "commongen.train.jsonl"
                  ),
                  "set_name": "train",
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.VALIDATION,
              gen_kwargs={
                  "filepath": os.path.join(
                      files["data"], "commongen.dev.jsonl"
                  ),
                  "set_name": "validation",
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.TEST,
              gen_kwargs={
                  "filepath": os.path.join(
                      files["data"], "commongen.test_noref.jsonl"
                  ),
                  "set_name": "test",
              },
          ),
      ] + challenge_splits

    elif self.builder_config.name == "cs_restaurants":
      challenge_sets = [
          (
              "challenge_train_sample",
              "train_cs_restaurants_RandomSample500.json",
          ),
          (
              "challenge_validation_sample",
              "validation_cs_restaurants_RandomSample500.json",
          ),
          (
              "challenge_test_scramble",
              "test_cs_restaurants_ScrambleInputStructure500.json",
          ),
      ]
      challenge_splits = []
      for challenge_split, filename in challenge_sets:
        challenge_splits.append(
            tfds.core.SplitGenerator(
                name=challenge_split,
                gen_kwargs={
                    "filepath": os.path.join(
                        files["challenge_set"],
                        self.builder_config.name,
                        filename,
                    ),
                    "set_name": challenge_split,
                },
            )
        )

      return [
          tfds.core.SplitGenerator(
              name=tfds.Split.TRAIN,
              gen_kwargs={
                  "filepath": files["train"],
                  "set_name": "train",
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.VALIDATION,
              gen_kwargs={
                  "filepath": files["validation"],
                  "set_name": "validation",
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.TEST,
              gen_kwargs={
                  "filepath": files["test"],
                  "set_name": "test",
              },
          ),
      ] + challenge_splits
    elif self.builder_config.name == "dart":
      return [
          tfds.core.SplitGenerator(
              name=tfds.Split.TRAIN,
              gen_kwargs={
                  "filepath": files["train"],
                  "set_name": "train",
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.VALIDATION,
              gen_kwargs={
                  "filepath": files["validation"],
                  "set_name": "validation",
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.TEST,
              gen_kwargs={
                  "filepath": files["test"],
                  "set_name": "test",
              },
          ),
      ]
    elif self.builder_config.name == "e2e_nlg":
      challenge_sets = [
          ("challenge_train_sample", "train_e2e_nlg_RandomSample500.json"),
          (
              "challenge_validation_sample",
              "validation_e2e_nlg_RandomSample500.json",
          ),
          (
              "challenge_test_scramble",
              "test_e2e_nlg_ScrambleInputStructure500.json",
          ),
      ]
      challenge_splits = []
      for challenge_split, filename in challenge_sets:
        challenge_splits.append(
            tfds.core.SplitGenerator(
                name=challenge_split,
                gen_kwargs={
                    "filepath": os.path.join(
                        files["challenge_set"],
                        self.builder_config.name,
                        filename,
                    ),
                    "set_name": challenge_split,
                },
            )
        )
      return [
          tfds.core.SplitGenerator(
              name=tfds.Split.TRAIN,
              gen_kwargs={
                  "filepath": files["train"],
                  "set_name": "train",
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.VALIDATION,
              gen_kwargs={
                  "filepath": files["validation"],
                  "set_name": "validation",
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.TEST,
              gen_kwargs={
                  "filepath": files["test"],
                  "set_name": "test",
              },
          ),
      ] + challenge_splits
    elif self.builder_config.name.startswith("mlsum"):
      # Can be either _de or _es.
      lang = self.builder_config.name.split("_")[1]
      challenge_sets = [
          (
              "challenge_train_sample",
              f"train_mlsum_{lang}_RandomSample500.json",
          ),
          (
              "challenge_validation_sample",
              f"validation_mlsum_{lang}_RandomSample500.json",
          ),
          ("challenge_test_covid", f"{lang}_test_covid19_cleaned.jsonl"),
      ]
      challenge_splits = []
      for challenge_split, filename in challenge_sets:
        challenge_splits.append(
            tfds.core.SplitGenerator(
                name=challenge_split,
                gen_kwargs={
                    "filepath": os.path.join(
                        files["challenge_set"],
                        self.builder_config.name,
                        filename,
                    ),
                    "set_name": challenge_split,
                },
            )
        )
      return [
          tfds.core.SplitGenerator(
              name=tfds.Split.TRAIN,
              gen_kwargs={
                  "filepath": os.path.join(
                      files["train"], lang + "_train.jsonl"
                  ),
                  "set_name": "train",
                  "lang": lang,
                  "filepaths": files["bad_ids"],
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.VALIDATION,
              gen_kwargs={
                  "filepath": os.path.join(
                      files["validation"], lang + "_val.jsonl"
                  ),
                  "set_name": "validation",
                  "lang": lang,
                  "filepaths": files["bad_ids"],
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.TEST,
              gen_kwargs={
                  "filepath": os.path.join(files["test"], lang + "_test.jsonl"),
                  "set_name": "test",
                  "lang": lang,
                  "filepaths": files["bad_ids"],
              },
          ),
      ] + challenge_splits
    elif self.builder_config.name == "schema_guided_dialog":
      challenge_sets = [
          (
              "challenge_train_sample",
              "train_schema_guided_dialog_RandomSample500_reformatted.json",
          ),
          (
              "challenge_validation_sample",
              "validation_schema_guided_dialog_RandomSample500_reformatted.json",
          ),
          (
              "challenge_test_backtranslation",
              "test_schema_guided_dialog_BackTranslation500_reformatted.json",
          ),
          (
              "challenge_test_bfp02",
              "test_schema_guided_dialog_ButterFingersPerturbation_p=0.02_500_reformatted.json",
          ),
          (
              "challenge_test_bfp05",
              "test_schema_guided_dialog_ButterFingersPerturbation_p=0.05_500_reformatted.json",
          ),
          (
              "challenge_test_nopunc",
              "test_schema_guided_dialog_WithoutPunctuation500_reformatted.json",
          ),
          (
              "challenge_test_scramble",
              "test_schema_guided_dialog_ScrambleInputStructure500_reformatted.json",
          ),
      ]
      challenge_splits = []
      for challenge_split, filename in challenge_sets:
        challenge_splits.append(
            tfds.core.SplitGenerator(
                name=challenge_split,
                gen_kwargs={
                    "filepath": os.path.join(
                        files["challenge_set"],
                        self.builder_config.name,
                        filename,
                    ),
                    "set_name": challenge_split,
                },
            )
        )

      generators = []
      for tfds_spl, spl in zip(
          [tfds.Split.TRAIN, tfds.Split.VALIDATION, tfds.Split.TEST],
          ["train", "validation", "test"],
      ):
        generators.append(
            tfds.core.SplitGenerator(
                name=tfds_spl,
                gen_kwargs={
                    "filepath": os.path.join(files["data"], "gem_sgd.json"),
                    "set_name": spl,
                },
            )
        )
      return generators + challenge_splits
    elif self.builder_config.name == "totto":
      challenge_sets = [
          ("challenge_train_sample", "train_totto_RandomSample500.json"),
          (
              "challenge_validation_sample",
              "validation_totto_RandomSample500.json",
          ),
          (
              "challenge_test_scramble",
              "test_totto_ScrambleInputStructure500.json",
          ),
      ]
      challenge_splits = []
      for challenge_split, filename in challenge_sets:
        challenge_splits.append(
            tfds.core.SplitGenerator(
                name=challenge_split,
                gen_kwargs={
                    "filepath": os.path.join(
                        files["challenge_set"],
                        self.builder_config.name,
                        filename,
                    ),
                    "set_name": challenge_split,
                },
            )
        )

      return [
          tfds.core.SplitGenerator(
              name=tfds.Split.TRAIN,
              gen_kwargs={
                  "filepath": os.path.join(
                      files["data"], "totto_data/totto_train_data.jsonl"
                  ),
                  "set_name": "train",
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.VALIDATION,
              gen_kwargs={
                  "filepath": os.path.join(
                      files["data"], "totto_data/totto_dev_data.jsonl"
                  ),
                  "set_name": "validation",
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.TEST,
              gen_kwargs={
                  "filepath": os.path.join(
                      files["data"],
                      "totto_data/unlabeled_totto_test_data.jsonl",
                  ),
                  "set_name": "test",
              },
          ),
      ] + challenge_splits
    elif self.builder_config.name.startswith("web_nlg"):
      # Can be either _en or _ru.
      ln = self.builder_config.name.split("_")[2]
      challenge_sets = [
          (
              "challenge_train_sample",
              f"train_web_nlg_{ln}_RandomSample500.json",
          ),
          (
              "challenge_validation_sample",
              f"validation_web_nlg_{ln}_RandomSample500.json",
          ),
          (
              "challenge_test_scramble",
              f"test_web_nlg_{ln}_ScrambleInputStructure500.json",
          ),
      ]
      if ln == "en":
        challenge_sets += [(
            "challenge_test_numbers",
            f"test_web_nlg_{ln}_replace_numbers_500.json",
        )]
      challenge_splits = []
      for challenge_split, filename in challenge_sets:
        challenge_splits.append(
            tfds.core.SplitGenerator(
                name=challenge_split,
                gen_kwargs={
                    "filepath": os.path.join(
                        files["challenge_set"],
                        self.builder_config.name,
                        filename,
                    ),
                    "set_name": challenge_split,
                },
            )
        )
      return [
          tfds.core.SplitGenerator(
              name=tfds.Split.TRAIN,
              gen_kwargs={"filepath": files["train"], "set_name": "train"},
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.VALIDATION,
              gen_kwargs={
                  "filepath": files["validation"],
                  "set_name": "validation",
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.TEST,
              gen_kwargs={"filepath": files["test"], "set_name": "test"},
          ),
      ] + challenge_splits
    elif self.builder_config.name == "wiki_auto_asset_turk":
      challenge_sets = [
          (
              "challenge_train_sample",
              "train_wiki_auto_asset_turk_RandomSample500.json",
          ),
          (
              "challenge_validation_sample",
              "validation_wiki_auto_asset_turk_RandomSample500.json",
          ),
          (
              "challenge_test_asset_backtranslation",
              "test_asset_wiki_auto_asset_turk_BackTranslation.json",
          ),
          (
              "challenge_test_asset_bfp02",
              "test_asset_wiki_auto_asset_turk_ButterFingersPerturbation_p=0.02.json",
          ),
          (
              "challenge_test_asset_bfp05",
              "test_asset_wiki_auto_asset_turk_ButterFingersPerturbation_p=0.05.json",
          ),
          (
              "challenge_test_asset_nopunc",
              "test_asset_wiki_auto_asset_turk_WithoutPunctuation.json",
          ),
          (
              "challenge_test_turk_backtranslation",
              "detok_test_turk_wiki_auto_asset_turk_BackTranslation.json",
          ),
          (
              "challenge_test_turk_bfp02",
              "detok_test_turk_wiki_auto_asset_turk_ButterFingersPerturbation_p=0.02.json",
          ),
          (
              "challenge_test_turk_bfp05",
              "detok_test_turk_wiki_auto_asset_turk_ButterFingersPerturbation_p=0.05.json",
          ),
          (
              "challenge_test_turk_nopunc",
              "detok_test_turk_wiki_auto_asset_turk_WithoutPunctuation.json",
          ),
      ]
      challenge_splits = []
      for challenge_split, filename in challenge_sets:
        challenge_splits.append(
            tfds.core.SplitGenerator(
                name=challenge_split,
                gen_kwargs={
                    "filepath": os.path.join(
                        files["challenge_set"],
                        self.builder_config.name,
                        filename,
                    ),
                    "set_name": challenge_split,
                },
            )
        )
      return [
          tfds.core.SplitGenerator(
              name=tfds.Split.TRAIN,
              gen_kwargs={
                  "filepath": files["train"],
                  "set_name": "train",
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.VALIDATION,
              gen_kwargs={
                  "filepath": files["validation"],
                  "set_name": "validation",
              },
          ),
          tfds.core.SplitGenerator(
              name="test_asset",
              gen_kwargs={
                  "filepath": "",
                  "set_name": "test_asset",
                  "filepaths": [
                      files["test_asset_" + str(i)] for i in range(10)
                  ],
              },
          ),
          tfds.core.SplitGenerator(
              name="test_turk",
              gen_kwargs={
                  "filepath": files["test_turk"],
                  "set_name": "test_turk",
              },
          ),
      ] + challenge_splits
    elif self.builder_config.name.startswith("wiki_lingua"):
      lang_name = self.builder_config.name.split("_")[-2]
      lang = self.builder_config.name.split("_")[-1]
      base_dir = os.path.join(files["data"], lang_name)
      return [
          tfds.core.SplitGenerator(
              name=tfds.Split.TRAIN,
              gen_kwargs={
                  "filepath": base_dir,
                  "set_name": "train",
                  "lang": lang,
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.VALIDATION,
              gen_kwargs={
                  "filepath": base_dir,
                  "set_name": "val",
                  "lang": lang,
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.TEST,
              gen_kwargs={
                  "filepath": base_dir,
                  "set_name": "test",
                  "lang": lang,
              },
          ),
      ]
    elif self.builder_config.name == "xsum":
      challenge_sets = [
          ("challenge_train_sample", "train_xsum_RandomSample500.json"),
          (
              "challenge_validation_sample",
              "validation_xsum_RandomSample500.json",
          ),
          (
              "challenge_test_backtranslation",
              "test_xsum_BackTranslation500.json",
          ),
          (
              "challenge_test_bfp_02",
              "test_xsum_ButterFingersPerturbation_p=0.02_500.json",
          ),
          (
              "challenge_test_bfp_05",
              "test_xsum_ButterFingersPerturbation_p=0.05_500.json",
          ),
          ("challenge_test_nopunc", "test_xsum_WithoutPunctuation500.json"),
          ("challenge_test_covid", "en_test_covid19.jsonl"),
      ]
      challenge_splits = []
      for challenge_split, filename in challenge_sets:
        challenge_splits.append(
            tfds.core.SplitGenerator(
                name=challenge_split,
                gen_kwargs={
                    "filepath": os.path.join(
                        files["challenge_set"],
                        self.builder_config.name,
                        filename,
                    ),
                    "set_name": challenge_split,
                },
            )
        )
      return [
          tfds.core.SplitGenerator(
              name=tfds.Split.TRAIN,
              gen_kwargs={
                  "filepath": files["splits"],
                  "set_name": "train",
                  "filepaths": os.path.join(files["data"], "bbc-summary-data"),
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.VALIDATION,
              gen_kwargs={
                  "filepath": files["splits"],
                  "set_name": "validation",
                  "filepaths": os.path.join(files["data"], "bbc-summary-data"),
              },
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.TEST,
              gen_kwargs={
                  "filepath": files["splits"],
                  "set_name": "test",
                  "filepaths": os.path.join(files["data"], "bbc-summary-data"),
              },
          ),
      ] + challenge_splits

  def _generate_examples(self, filepath, set_name, filepaths=None, lang=None):
    """Yields examples."""
    if self.builder_config.name == "common_gen":
      with epath.Path(filepath).open() as f:
        if set_name.startswith("challenge"):
          exples = json.load(f)
          if isinstance(exples, dict):
            assert len(exples) == 1, "multiple entries found"
            exples = list(exples.values())[0]
          for id_, exple in enumerate(exples):
            if not exple:
              continue
            exple["gem_parent_id"] = exple["gem_id"]
            exple["gem_id"] = f"{self.builder_config.name}-{set_name}-{id_}"
            yield id_, exple
        else:
          id_ = -1
          i = -1
          for row in f:
            row = row.replace(", }", "}")  # Fix possible JSON format error.
            data = json.loads(row)
            concepts = [word for word in data["concept_set"].split("#")]
            if set_name == "train":
              i += 1
              for scene in data["scene"]:
                id_ += 1
                yield id_, {
                    "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
                    "gem_parent_id": (
                        f"{self.builder_config.name}-{set_name}-{id_}"
                    ),
                    "concept_set_id": i,
                    "concepts": concepts,
                    "target": scene,
                    "references": [],
                }
            else:
              id_ += 1
              yield id_, {
                  "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
                  "gem_parent_id": (
                      f"{self.builder_config.name}-{set_name}-{id_}"
                  ),
                  "concept_set_id": id_,
                  "concepts": concepts,
                  "target": "" if set_name == "test" else data["scene"][0],
                  "references": [] if set_name == "test" else data["scene"],
              }
    elif self.builder_config.name == "cs_restaurants":
      with epath.Path(filepath).open() as f:
        if set_name.startswith("challenge"):
          exples = json.load(f)
          if isinstance(exples, dict):
            assert len(exples) == 1, "multiple entries found"
            exples = list(exples.values())[0]
          for id_, exple in enumerate(exples):
            if not exple:
              continue
            exple["gem_parent_id"] = exple["gem_id"]
            exple["gem_id"] = f"{self.builder_config.name}-{set_name}-{id_}"
            yield id_, exple
        else:
          data = json.load(f)
          for id_, instance in enumerate(data):
            yield id_, {
                "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
                "gem_parent_id": f"{self.builder_config.name}-{set_name}-{id_}",
                "dialog_act": instance["da"],
                "dialog_act_delexicalized": instance["delex_da"],
                "target": instance["text"],
                "target_delexicalized": instance["delex_text"],
                "references": [] if set_name == "train" else [instance["text"]],
            }
    elif self.builder_config.name == "dart":
      with epath.Path(filepath).open() as f:
        data = json.loads(f.read())
        id_ = -1
        i = -1
        for example in data:
          flat_tripleset = [" ".join(ex) for ex in example["tripleset"]]
          if set_name == "train":
            i += 1
            for annotation in example["annotations"]:
              id_ += 1
              yield id_, {
                  "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
                  "gem_parent_id": (
                      f"{self.builder_config.name}-{set_name}-{id_}"
                  ),
                  "dart_id": i,
                  "tripleset": flat_tripleset,
                  "subtree_was_extended": example.get(
                      "subtree_was_extended", None
                  ),  # some are missing.
                  "target_sources": [
                      annotation["source"]
                      for annotation in example["annotations"]
                  ],
                  "target": annotation["text"],
                  "references": [],
              }
          else:
            id_ += 1
            yield id_, {
                "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
                "gem_parent_id": f"{self.builder_config.name}-{set_name}-{id_}",
                "dart_id": id_,
                "tripleset": flat_tripleset,
                "subtree_was_extended": example.get(
                    "subtree_was_extended", None
                ),  # some are missing.
                "target_sources": [
                    annotation["source"]
                    for annotation in example["annotations"]
                ],
                "target": (
                    example["annotations"][0]["text"]
                    if example["annotations"]
                    else ""
                ),
                "references": [
                    annotation["text"] for annotation in example["annotations"]
                ],
            }
    elif self.builder_config.name == "e2e_nlg":
      with epath.Path(filepath).open() as f:
        if set_name.startswith("challenge"):
          exples = json.load(f)
          if isinstance(exples, dict):
            assert len(exples) == 1, "multiple entries found"
            exples = list(exples.values())[0]
          for id_, exple in enumerate(exples):
            if not exple:
              continue
            exple["gem_parent_id"] = exple["gem_id"]
            exple["gem_id"] = f"{self.builder_config.name}-{set_name}-{id_}"
            yield id_, exple
        else:
          reader = csv.DictReader(f)
          for id_, example in enumerate(reader):
            yield id_, {
                "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
                "gem_parent_id": f"{self.builder_config.name}-{set_name}-{id_}",
                "meaning_representation": example["mr"],
                "target": example["ref"],
                "references": [] if set_name == "train" else [example["ref"]],
            }
    elif self.builder_config.name.startswith("mlsum"):
      if set_name in ["train", "validation", "test", "challenge_test_covid"]:
        if set_name == "challenge_test_covid":
          bad_ids = {}
        else:
          bad_ids_dct = json.load(tf.io.gfile.GFile(filepaths))
          bad_ids = dict(
              (bad_url, True)
              for _, bad_url in bad_ids_dct[f"{lang}-{set_name}"]
          )
        with epath.Path(filepath).open() as f:
          id_ = -1
          for line in f:
            data = json.loads(line)
            if data["url"] in bad_ids:
              continue
            else:
              id_ += 1
              yield id_, {
                  "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
                  "gem_parent_id": (
                      f"{self.builder_config.name}-{set_name}-{id_}"
                  ),
                  "text": data["text"],
                  "target": data["summary"],
                  "references": (
                      [] if set_name == "train" else [data["summary"]]
                  ),
                  "topic": data["topic"],
                  "url": data["url"],
                  "title": data["title"],
                  "date": data["date"],
              }
      else:
        exples = json.load(tf.io.gfile.GFile(filepath))
        if isinstance(exples, dict):
          assert len(exples) == 1, "multiple entries found"
          exples = list(exples.values())[0]
        for id_, exple in enumerate(exples):
          if not exple:
            continue
          exple["gem_parent_id"] = exple["gem_id"]
          exple["gem_id"] = f"{self.builder_config.name}-{set_name}-{id_}"
          yield id_, exple
    elif self.builder_config.name == "schema_guided_dialog":
      if "challenge" in set_name:
        exples = json.load(tf.io.gfile.GFile(filepath))
        if isinstance(exples, dict):
          assert len(exples) == 1, "multiple entries found"
          exples = list(exples.values())[0]
        for id_, exple in enumerate(exples):
          if not exple:
            continue
          exple["gem_parent_id"] = exple["gem_id"]
          exple["gem_id"] = f"{self.builder_config.name}-{set_name}-{id_}"
          yield id_, exple
      else:
        examples = json.load(tf.io.gfile.GFile(filepath))[set_name]
        for id_, example in enumerate(examples):
          dialog_acts = []
          for act_id, slot, values in example["da"]:
            dialog_acts.append({
                "act": act_id,
                "slot": slot,
                "values": values,
            })
          yield id_, {
              "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
              "gem_parent_id": f"{self.builder_config.name}-{set_name}-{id_}",
              "dialog_acts": dialog_acts,
              "context": example["context"],
              "dialog_id": example["dialog_id"],
              "service": example["service"],
              "turn_id": example["turn_ix"],
              "prompt": example["prompt"],
              "target": example["target"],
              "references": [] if set_name == "train" else [example["target"]],
          }
    elif self.builder_config.name == "totto":
      if "challenge" in set_name:
        exples = json.load(tf.io.gfile.GFile(filepath))
        if isinstance(exples, dict):
          assert len(exples) == 1, "multiple entries found"
          exples = list(exples.values())[0]
        for id_, exple in enumerate(exples):
          if not exple:
            continue
          exple["gem_parent_id"] = exple["gem_id"]
          exple["gem_id"] = f"{self.builder_config.name}-{set_name}-{id_}"
          yield id_, exple
      else:
        with epath.Path(filepath).open() as json_file:
          json_list = list(json_file)
        id_ = -1
        i = -1
        for json_str in json_list:
          result = json.loads(json_str)
          if set_name == "train":
            i += 1
            for sentence in result["sentence_annotations"]:
              id_ += 1
              response = {
                  "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
                  "gem_parent_id": (
                      f"{self.builder_config.name}-{set_name}-{id_}"
                  ),
                  "totto_id": i,
                  "table_page_title": result["table_page_title"],
                  "table_webpage_url": result["table_webpage_url"],
                  "table_section_title": result["table_section_title"],
                  "table_section_text": result["table_section_text"],
                  "table": result["table"],
                  "highlighted_cells": result["highlighted_cells"],
                  "example_id": str(result["example_id"]),
                  "overlap_subset": "none",
                  "sentence_annotations": [sentence],
                  "references": [],
                  "target": sentence["final_sentence"],
              }
              yield id_, response
          else:
            id_ += 1
            response = {
                "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
                "gem_parent_id": f"{self.builder_config.name}-{set_name}-{id_}",
                "totto_id": id_,
                "table_page_title": result["table_page_title"],
                "table_webpage_url": result["table_webpage_url"],
                "table_section_title": result["table_section_title"],
                "table_section_text": result["table_section_text"],
                "table": result["table"],
                "highlighted_cells": result["highlighted_cells"],
                "example_id": str(result["example_id"]),
                "overlap_subset": str(result["overlap_subset"]),
            }
            response["sentence_annotations"] = (
                [] if set_name == "test" else result["sentence_annotations"]
            )
            response["references"] = [
                sentence["final_sentence"]
                for sentence in response["sentence_annotations"]
            ]
            if response["references"]:
              response["target"] = response["references"][0]
            else:
              response["target"] = ""
            yield id_, response
    elif self.builder_config.name.startswith("web_nlg"):
      if "challenge" in set_name:
        exples = json.load(tf.io.gfile.GFile(filepath))
        if isinstance(exples, dict):
          assert len(exples) == 1, "multiple entries found"
          exples = list(exples.values())[0]
        for id_, exple in enumerate(exples):
          if not exple:
            continue
          exple["gem_parent_id"] = exple["gem_id"]
          exple["gem_id"] = f"{self.builder_config.name}-{set_name}-{id_}"
          yield id_, exple
      else:
        with epath.Path(filepath).open() as f:
          examples = json.load(f)
          id_ = -1
          for example in examples["values"]:
            if set_name == "train":
              for target in example["target"]:
                id_ += 1
                yield id_, {
                    "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
                    "gem_parent_id": (
                        f"{self.builder_config.name}-{set_name}-{id_}"
                    ),
                    "input": example["input"],
                    "target": target,
                    "references": (
                        [] if set_name == "train" else example["target"]
                    ),
                    "category": example["category"],
                    "webnlg_id": example["webnlg-id"],
                }
            else:
              id_ += 1
              yield id_, {
                  "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
                  "gem_parent_id": (
                      f"{self.builder_config.name}-{set_name}-{id_}"
                  ),
                  "input": example["input"],
                  "target": example["target"][0] if example["target"] else "",
                  "references": example["target"],
                  "category": example["category"],
                  "webnlg_id": example["webnlg-id"],
              }
    elif self.builder_config.name == "wiki_auto_asset_turk":
      if set_name in ["train", "validation"]:
        keys = [
            "source",
            "target",
        ]
        with epath.Path(filepath).open() as f:
          for id_, line in enumerate(f):
            values = line.strip().split("\t")
            assert (
                len(values) == 2
            ), f"Not enough fields in ---- {line} --- {values}"
            example = dict([(k, val) for k, val in zip(keys, values)])
            example["gem_id"] = f"{self.builder_config.name}-{set_name}-{id_}"
            example["gem_parent_id"] = (
                f"{self.builder_config.name}-{set_name}-{id_}"
            )
            example["references"] = (
                [] if set_name == "train" else [example["target"]]
            )
            yield id_, example
      elif set_name == "test_turk":
        examples = json.load(tf.io.gfile.GFile(filepath))
        for id_, example in enumerate(examples):
          example["gem_parent_id"] = example["gem_id"]
          for k in ["source_id", "target_id"]:
            if k in example:
              del example[k]
          yield id_, example
      elif set_name == "test_asset":
        files = [tf.io.gfile.GFile(f_name) for f_name in filepaths]
        for id_, lines in enumerate(zip(*files)):
          yield id_, {
              "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
              "gem_parent_id": f"{self.builder_config.name}-{set_name}-{id_}",
              "target": lines[1].strip(),
              "source": lines[0].strip(),
              "references": [line.strip() for line in lines[1:]],
          }
      else:
        exples = json.load(tf.io.gfile.GFile(filepath))
        if isinstance(exples, dict):
          assert len(exples) == 1, "multiple entries found"
          exples = list(exples.values())[0]
        for id_, exple in enumerate(exples):
          exple["gem_parent_id"] = exple["gem_id"]
          exple["gem_id"] = f"{self.builder_config.name}-{set_name}-{id_}"
          for k in ["source_id", "target_id"]:
            if k in exple:
              del exple[k]
          yield id_, exple
    elif self.builder_config.name.startswith("wiki_lingua"):
      with tf.io.gfile.GFile(
          os.path.join(filepath, f"{set_name}.src.{lang}")
      ) as f_in_ln:
        with tf.io.gfile.GFile(
            os.path.join(filepath, f"{set_name}.src.en")
        ) as f_in_en:
          with tf.io.gfile.GFile(
              os.path.join(filepath, f"{set_name}.tgt.{lang}")
          ) as f_out_ln:
            with tf.io.gfile.GFile(
                os.path.join(filepath, f"{set_name}.tgt.en")
            ) as f_out_en:
              for id_, (src_ln, src_en, tgt_ln, tgt_en) in enumerate(
                  zip(f_in_ln, f_in_en, f_out_ln, f_out_en)
              ):
                yield id_, {
                    "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
                    "gem_parent_id": (
                        f"{self.builder_config.name}-{set_name}-{id_}"
                    ),
                    "source_aligned": {
                        lang: src_ln.strip(),
                        "en": src_en.strip(),
                    },
                    "target_aligned": {
                        lang: tgt_ln.strip(),
                        "en": tgt_en.strip(),
                    },
                    "source": src_ln.strip(),
                    "target": tgt_en.strip(),
                    "references": (
                        [] if set_name == "train" else [tgt_en.strip()]
                    ),
                }
    elif self.builder_config.name == "xsum":
      if "challenge" in set_name:
        if "covid" in set_name:
          with epath.Path(filepath).open() as f:
            id_ = -1
            for line in f:
              data = json.loads(line)
              id_ += 1
              yield id_, {
                  "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
                  "gem_parent_id": (
                      f"{self.builder_config.name}-{set_name}-{id_}"
                  ),
                  "xsum_id": data["url"],
                  "document": data["text"],
                  "target": data["summary"],
                  "references": (
                      [] if set_name == "train" else [data["summary"]]
                  ),
              }
        else:
          exples = json.load(tf.io.gfile.GFile(filepath))
          if isinstance(exples, dict):
            assert len(exples) == 1, "multiple entries found"
            exples = list(exples.values())[0]
          for id_, exple in enumerate(exples):
            exple["gem_parent_id"] = exple["gem_id"]
            exple["gem_id"] = f"{self.builder_config.name}-{set_name}-{id_}"
            yield id_, exple
      else:
        with epath.Path(filepath).open() as f:
          split_ids = json.load(f)
        for id_, i in enumerate(split_ids[set_name]):
          with tf.io.gfile.GFile(os.path.join(filepaths, i + ".summary")) as f:
            text = "".join(
                [
                    line
                    for line in f.readlines()
                    if line not in _XSUM_REMOVE_LINES and line.strip()
                ]
            )
            segs = text.split("[SN]")
            yield id_, {
                "gem_id": f"{self.builder_config.name}-{set_name}-{id_}",
                "gem_parent_id": f"{self.builder_config.name}-{set_name}-{id_}",
                "xsum_id": i,
                "document": segs[8].strip(),
                "target": segs[6].strip(),
                "references": [] if set_name == "train" else [segs[6].strip()],
            }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils to generate builders for D4RL datasets."""
from typing import Any, Dict

import h5py
import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

_DESCRIPTION = """
D4RL is an open-source benchmark for offline reinforcement learning. It provides
standardized environments and datasets for training and benchmarking algorithms.

The datasets follow the [RLDS format](https://github.com/google-research/rlds)
to represent steps and episodes.
"""

_CITATION = """\
@misc{fu2020d4rl,
    title={D4RL: Datasets for Deep Data-Driven Reinforcement Learning},
    author={Justin Fu and Aviral Kumar and Ofir Nachum and George Tucker and Sergey Levine},
    year={2020},
    eprint={2004.07219},
    archivePrefix={arXiv},
    primaryClass={cs.LG}
}
"""


def description():
  return _DESCRIPTION


def citation():
  return _CITATION


def url():
  return 'https://sites.google.com/view/d4rl/home'


def generate_examples(file_path: str):
  """Provides a common generate_examples method for D4RL datasets."""
  d4rl_dict = read_d4rl_dataset(file_path)
  if 'timeouts' not in d4rl_dict:
    raise ValueError('Only datasets with explicit timeouts are supported.')

  done = [
      terminal or timeout
      for (terminal, timeout) in zip(
          d4rl_dict['terminals'], d4rl_dict['timeouts']
      )
  ]
  # is_first corresponds to the done flag delayed by one step.
  d4rl_dict['is_first'] = [True] + done[:-1]
  # is_last is not used but this is needed to build a valid dictionary.
  d4rl_dict['is_last'] = done

  # Get step metadata
  infos_dict = _get_nested_metadata(d4rl_dict, 'infos')

  # Flatten reward
  d4rl_dict['rewards'] = np.squeeze(d4rl_dict['rewards'])

  episode_metadata = _get_nested_metadata(d4rl_dict, 'metadata')
  dataset_dict = {
      'observation': d4rl_dict['observations'],
      'action': d4rl_dict['actions'],
      'reward': d4rl_dict['rewards'],
      'discount': np.ones_like(d4rl_dict['rewards']),
      'is_terminal': d4rl_dict['terminals'],
      'is_first': d4rl_dict['is_first'],
      'is_last': d4rl_dict['is_last'],
  }
  if 'next_observations' in d4rl_dict:
    dataset_dict['next_observation'] = d4rl_dict['next_observations']

  if infos_dict:
    dataset_dict['infos'] = infos_dict
  num_steps = len(dataset_dict['is_first'])
  prev = 0
  counter = 0
  for pos in range(num_steps):
    if dataset_dict['is_first'][pos] and pos > prev:
      yield counter, _get_episode(dataset_dict, episode_metadata, prev, pos)
      prev = pos
      counter += 1
  if prev < num_steps:
    yield counter, _get_episode(dataset_dict, episode_metadata, prev, num_steps)


def _get_nested_metadata(
    dataset: Dict[str, Any], prefix: str
) -> Dict[str, Any]:
  """Generate a metadata dictionary using flattened metadata keys.

  Args:
    dataset: dictionary containing the dataset keys and values. Keys are
      flatened.
    prefix: common prefix of the metadata fields.

  Returns:
    Nested dictionary with the episode metadata.

  If the dataset contains:
  {
    'metadata/v1/v2': 1,
    'metadata/v3': 2,
  }
  and prefix='metadata', it returns:
  {
    'v1':{
      'v2': 1,
    }
    'v3': 2,
  }
  It assumes that the flattened metadata keys are well-formed.
  """
  episode_metadata = {}
  for k in dataset.keys():
    if f'{prefix}/' not in k:
      continue
    keys = k.split('/')[1:]
    nested_dict = episode_metadata
    leaf_value = dataset[k]
    for index, nested_key in enumerate(keys):
      if index == (len(keys) - 1):
        nested_dict[nested_key] = leaf_value
      else:
        if nested_key not in nested_dict:
          nested_dict[nested_key] = {}
        nested_dict = nested_dict[nested_key]

  return episode_metadata


def _get_episode(
    steps: Dict[str, Any],
    episode_metadata: Dict[str, Any],
    begin: int,
    end: int,
) -> Dict[str, Any]:
  """Builds a full episode dict.

  Args:
      steps: a dict with all steps in a dataset
      episode_metadata: dict with the episode metadata
      begin: defines a starting position of an episode
      end: defines an ending position of an episode

  Returns:
     A dict with data specific to one episode, already broken into steps.
  """
  # It's an initial step if the episode is empty.
  episode = {}
  for k in [
      'is_first',
      'is_last',
      'observation',
      'action',
      'reward',
      'discount',
  ]:
    episode[k] = steps[k][begin:end]

  episode['is_last'] = [False] * (end - begin)
  episode['is_terminal'] = [False] * (end - begin)
  if 'infos' in steps.keys():
    episode['infos'] = {}
    for k in steps['infos'].keys():
      episode['infos'][k] = steps['infos'][k][begin:end]

  ends_in_terminal = steps['is_terminal'][end - 1]
  has_next_obs = 'next_observation' in steps

  # If the episode ends in a terminal state, the discount of the previous step
  # is set to 0 (it's a transition to a terminal state).
  if ends_in_terminal:
    episode['discount'][-1] = 0.0
  if ends_in_terminal or has_next_obs:
    # In HDF5, the terminal bit is associated with the previous observation.
    # To comply with RLDS standard (see types.py), we propagate this information
    # to a next state. This matches the definition in RLDS. See types.py.
    # If there is information about a next observation, we also generate an
    # extra last step.
    episode['is_first'] = np.concatenate((episode['is_first'], [False]))
    if has_next_obs:
      episode['observation'] = np.concatenate(
          (episode['observation'], [steps['next_observation'][-1]])
      )
    else:
      # If the last observation is never recorded and to avoid discarding the
      # last transition to the terminal state, we create a dummy observation.
      # Since no solution is perfect, the design choice was to keep as much
      # information as possible and let the user decide to keep or ignore such
      # transitions.
      episode['observation'] = np.concatenate(
          (episode['observation'], [np.zeros_like(steps['observation'][0])])
      )
    # Action/reward/discount are set to dummy values since not relevant.
    # When IS_LAST is set, any field coming temporally after the last
    # observation is invalid.
    episode['action'] = np.concatenate(
        (episode['action'], [np.zeros_like(steps['action'][0])])
    )
    episode['reward'] = np.concatenate(
        (episode['reward'], [np.zeros_like(steps['reward'][0])])
    )
    episode['discount'] = np.array(
        np.concatenate((episode['discount'], [0.0])), dtype=np.float32
    )

    episode['is_terminal'] = np.concatenate(
        (episode['is_terminal'], [ends_in_terminal])
    )
    episode['is_last'] = np.concatenate((episode['is_last'], [True]))
    if 'infos' in steps.keys():
      for k in steps['infos'].keys():
        episode['infos'][k] = np.concatenate(
            (episode['infos'][k], [np.zeros_like(steps['infos'][k][0])])
        )
  else:
    # Despite the fact that the last action and reward are valid in the
    # stored dataset (in the final transition, [obs, action, reward, next_obs],
    # only one step was stored as [obs, action, reward], instead of adding one
    # extra step with [next_obs, 0, 0]). We set IS_LAST=True so that
    # it is consistent with other typical datasets.
    episode['is_last'][-1] = True
  full_episode = {'steps': episode}
  if episode_metadata:
    full_episode.update(episode_metadata)
  return full_episode


def _get_dataset_keys(h5file):
  """Gets the keys present in the D4RL dataset."""
  keys = []

  def visitor(name, item):
    if isinstance(item, h5py.Dataset):
      keys.append(name)

  h5file.visititems(visitor)
  return keys


def read_d4rl_dataset(file_path: str):
  """Reads a D4RL dataset and returns the dataset as a dictionary."""
  with tf.io.gfile.GFile(file_path, 'rb') as f:
    with h5py.File(f, 'r') as dataset_file:
      dataset_dict = {}
      for k in _get_dataset_keys(dataset_file):
        try:
          # first try loading as an array
          dataset_dict[k] = dataset_file[k][:]
        except ValueError:  # try loading as a scalar
          dataset_dict[k] = dataset_file[k][()]

    return dataset_dict



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pytest plugins globally available.

`conftest.py` file is automatically detected by Pytest and register
plugins (hooks and fixtures) common to all tests.

See: https://docs.pytest.org/en/latest/writing_plugins.html
"""
from __future__ import annotations

import builtins
import importlib
import sys
import typing
from typing import Iterator, Type

import pytest
from tensorflow_datasets import setup_teardown

if typing.TYPE_CHECKING:
  from tensorflow_datasets import testing
  from tensorflow_datasets.core import dataset_builder

# Global setup/teardown

# Ignore tests and non-test files
# https://docs.pytest.org/en/latest/reference/reference.html#globalvar-collect_ignore
collect_ignore = ['testing/test_utils.py', 'core/features/features_test.py']

# Register fixtures which are automatically applied once when tests start.


@pytest.fixture(scope='session', autouse=True)
def disable_community_datasets():
  """During tests, `tfds.list_builders` disable community datasets."""
  # For tests, only public datasets are available (no-community datasets)
  # Kokoro pytest tests are executed without absl.app, so the default
  # visibility isn't automatically set.
  from tensorflow_datasets.core import visibility  # pylint: disable=g-import-not-at-top

  visibility.set_availables(
      [
          visibility.DatasetType.TFDS_PUBLIC,
      ]
  )


@pytest.fixture(scope='session', autouse=True)
def skip_test_if_array_record_not_supported():
  """Skip all tests that depend on `array_record` package if it's not supported."""
  if sys.platform == 'linux':
    yield
  else:

    def patch(func):
      def patched_func(*args, **kwargs):
        name = kwargs.get('name', args[0])
        if name.startswith('array_record'):
          pytest.skip(
              reason=(
                  'ArrayRecord file format is not supported on your platform,'
                  ' skipping affected tests.'
              ),
          )
        return func(*args, **kwargs)

      return patched_func

    with pytest.MonkeyPatch.context() as monkeypatch:
      monkeypatch.setattr('builtins.__import__', patch(builtins.__import__))
      monkeypatch.setattr(
          'importlib.import_module', patch(importlib.import_module)
      )
      yield monkeypatch


# Register all fixtures defined in `setup_teardown` to be automatically
# applied in all tests.
global_dict = globals()
for fixture_fn in setup_teardown.GLOBAL_FIXTURES:
  fixture_name = fixture_fn.__name__
  if fixture_name in global_dict:
    raise ValueError(f'{fixture_name} already in module.')
  fixture_fn = pytest.fixture(scope='session', autouse=True)(fixture_fn)
  # In orders for fixtures to be registered, there need to be an explicit
  # attribute
  # https://stackoverflow.com/questions/27064004/splitting-a-conftest-py-file-into-several-smaller-conftest-like-parts/65035367#65035367
  global_dict[fixture_name] = fixture_fn
del global_dict  # Do not modifying global beyond this point

# Fixtures globally available


@pytest.fixture
def mock_fs() -> Iterator[testing.MockFs]:
  """Patch `tf.io.gfile` API into a virtual file system."""
  from tensorflow_datasets import testing  # pylint: disable=g-import-not-at-top

  with testing.MockFs() as fs:
    yield fs


def _make_dataset(
    tmp_path_factory: pytest.TempPathFactory,
    builder_cls: Type[dataset_builder.DatasetBuilder],
) -> dataset_builder.DatasetBuilder:
  tmp_path = tmp_path_factory.mktemp(f'global_{builder_cls.__name__}')
  builder = builder_cls(data_dir=tmp_path)
  builder.download_and_prepare()
  return builder


@pytest.fixture(scope='session')
def dummy_mnist(
    tmp_path_factory: pytest.TempPathFactory,
) -> dataset_builder.DatasetBuilder:
  """Dummy mnist dataset builder pre-generated."""
  from tensorflow_datasets import testing  # pylint: disable=g-import-not-at-top

  return _make_dataset(tmp_path_factory, testing.DummyMnist)


@pytest.fixture(scope='session')
def dummy_dataset(
    tmp_path_factory: pytest.TempPathFactory,
) -> dataset_builder.DatasetBuilder:
  """Dummy dataset builder pre-generated."""
  from tensorflow_datasets import testing  # pylint: disable=g-import-not-at-top

  return _make_dataset(tmp_path_factory, testing.DummyDataset)





# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for Winogrande dataset module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow_datasets import testing
from tensorflow_datasets.text import winogrande


class WinograndeTest(testing.DatasetBuilderTestCase):
  DATASET_CLASS = winogrande.Winogrande

  SPLITS = {
      "train_xl": 2,
      "train_l": 2,
      "train_m": 2,
      "train_s": 2,
      "train_xs": 2,
      "validation": 2,
      "test": 2,
  }

  OVERLAPPING_SPLITS = [
      "train_{}".format(size) for size in ("xs", "s", "m", "l", "xl")
  ]


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Re-compiles tf.train.Example and tf.train.Feature protos from the source.

Disclaimer: we copy tf.train.Example and tf.train.Feature from TensorFlow.

We did this in order not to import all the TensorFlow library when we only need
these two protos.

Execute this script with a local version of Python (e.g., using `pyenv`):

```
python tensorflow_datasets/proto/build_tf_proto.py
```
"""

import enum
import os
import re
import subprocess

from absl import app
from etils import epath
import requests

PATH_TO_TFDS = epath.Path(__file__).parent.parent
PATH_TO_PROTO = PATH_TO_TFDS / 'proto'


class TFProto(enum.Enum):
  EXAMPLE = 'example'
  FEATURE = 'feature'


def get_tf_proto_content_from_github(tf_proto: TFProto) -> bytes:
  """Downloads the proto from Github and formats it."""
  url = f'https://raw.githubusercontent.com/tensorflow/tensorflow/master/tensorflow/core/example/{tf_proto.value}.proto'
  request = requests.get(url)
  content = request.content
  content = clean_import_in_proto(
      content,
      b'import "tensorflow/core/example/feature.proto";',
      b'import "tensorflow_datasets/proto/tf_feature.proto";',
  )
  content = clean_package(content)
  content = remove_comments(content)
  content = add_header(content, url)
  return content


def get_tfds_proto_path(tf_proto: TFProto) -> epath.Path:
  return PATH_TO_PROTO / f'tf_{tf_proto.value}.proto'


def add_header(file_content: bytes, url: str) -> bytes:
  file_content_with_header = b'// DO NOT EDIT\n'
  file_content_with_header += b'// This file was copied from\n'
  file_content_with_header += b'// ' + url.encode() + b'\n\n'
  file_content_with_header += b'// Use according to the original license.\n\n'
  file_content_with_header += file_content
  return file_content_with_header


def clean_import_in_proto(
    file_content: bytes, before: bytes, after: bytes
) -> bytes:
  return file_content.replace(before, after)


def clean_package(file_content: bytes) -> bytes:
  """Declares a new package in order to avoid protobuf typing conflicts."""
  return file_content.replace(
      b'package tensorflow;',
      b'package tensorflow_copy;',
  )


def remove_comments(file_content: bytes) -> bytes:
  file_content = re.sub(rb'^\s*\/\/.*\n', b'', file_content, flags=re.MULTILINE)
  file_content = re.sub(rb'\s*\/\/.*\n', b'\n', file_content)
  return file_content


def change_imports_in_py(
    file_content: bytes, name_before: str, name_after: str
) -> bytes:
  name_before_without_suffix = os.path.splitext(name_before)[0]
  name_after_without_suffix = os.path.splitext(name_after)[0]
  return file_content.replace(
      f'import {name_before_without_suffix} as'.encode(),
      f'import {name_after_without_suffix} as'.encode(),
  )


def skip_pylint(file_content: bytes) -> bytes:
  file_content_without_pylint = b'# pylint: skip-file\n\n'
  file_content_without_pylint += file_content
  return file_content_without_pylint


def main(_) -> None:
  names_before = []
  names_after = []
  for proto_name in [TFProto.FEATURE, TFProto.EXAMPLE]:
    tf_proto_content = get_tf_proto_content_from_github(proto_name)
    tfds_proto_path = get_tfds_proto_path(proto_name)
    tfds_proto_path.write_bytes(tf_proto_content)

    # Launch protoc
    subprocess.run(
        [
            'protoc',
            os.fspath(tfds_proto_path),
            f'--proto_path={PATH_TO_TFDS.parent}',
            f'--python_out={PATH_TO_TFDS.parent}',
        ],
        check=True,
    )
    name_before = PATH_TO_PROTO / f'tf_{proto_name.value}_pb2.py'
    name_after = PATH_TO_PROTO / f'tf_{proto_name.value}_generated_pb2.py'
    names_before.append(name_before)
    names_after.append(name_after)

  # Rename tf_*_pb2.py -> tf_*_generated_pb2.py
  for name_before, name_after in zip(names_before, names_after):
    subprocess.run(
        ['mv', os.fspath(name_before), os.fspath(name_after)],
        check=True,
    )
    target_py_content = name_after.read_bytes()
    target_py_content = skip_pylint(target_py_content)
    for other_name_before, other_name_after in zip(names_before, names_after):
      target_py_content = change_imports_in_py(
          target_py_content, other_name_before.name, other_name_after.name
      )
    name_after.write_bytes(target_py_content)


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test of `document_datasets.py`."""

import functools
from typing import Set
from unittest import mock

import pytest

import tensorflow_datasets as tfds
from tensorflow_datasets.scripts.documentation import doc_utils
from tensorflow_datasets.scripts.documentation import document_datasets


class DummyDatasetConfigs(tfds.testing.DummyDataset):
  """Builder with config and manual instructions."""

  MANUAL_DOWNLOAD_INSTRUCTIONS = """Some manual instructions."""
  BUILDER_CONFIGS = [
      tfds.core.BuilderConfig(
          name='config_name',
          version=tfds.core.Version('0.0.1'),
          description='Config description.',
      ),
  ]


class DummyDatasetConfigsSharedVersion(tfds.testing.DummyDataset):
  """Builder with config ."""

  # No BuilderConfig description, and version shared across configs.
  VERSION = tfds.core.Version('1.0.0')
  BUILDER_CONFIGS = [
      tfds.core.BuilderConfig(name='config_name'),
  ]


@pytest.fixture
def document_single_builder_fn(tmp_path):
  yield functools.partial(
      document_datasets._document_single_builder,
      visu_doc_util=doc_utils.VisualizationDocUtil(
          base_path=tmp_path,
          base_url=doc_utils.DocUtilPaths.fig_base_url,
      ),
      df_doc_util=doc_utils.DataframeDocUtil(
          base_path=tmp_path,
          base_url=doc_utils.DocUtilPaths.df_base_url,
      ),
      nightly_doc_util=None,
  )


def test_document_datasets():
  all_docs = list(
      document_datasets.iter_documentation_builders(
          # Builder with and without config, as well as config-based builder.
          datasets=['mnist', 'coco', 'pass'],
          doc_util_paths=doc_utils.DocUtilPaths(
              fig_base_path=None,
              df_base_path=None,
              nightly_path=None,
          ),
      )
  )
  assert {d.name for d in all_docs} == {'mnist', 'coco', 'pass'}


def test_document_collection():
  all_docs = list(
      document_datasets.iter_collections_documentation(
          collection_names=['xtreme', 'longt5']
      )
  )
  assert {d.name for d in all_docs} == {'xtreme', 'longt5'}


def test_with_config(document_single_builder_fn):  # pylint: disable=redefined-outer-name
  """Test that builder with configs are correctly generated."""
  doc = document_single_builder_fn(DummyDatasetConfigs.name)
  assert 'Some manual instructions.' in doc.content
  assert 'Minimal DatasetBuilder.' in doc.content  # Shared description.
  # Config-description
  assert '**Config description**: Config description.' in doc.content
  assert (
      '<meta itemprop="url" content="'
      f'https://www.tensorflow.org/datasets/catalog/{DummyDatasetConfigs.name}"'
      ' />'
      in doc.content
  )


def test_with_config_shared_version(document_single_builder_fn):  # pylint: disable=redefined-outer-name
  """Test that builder with configs are correctly generated."""
  doc = document_single_builder_fn(DummyDatasetConfigsSharedVersion.name)
  assert 'Minimal DatasetBuilder.' in doc.content  # Shared description.
  assert 'Config description:' not in doc.content  # No config description.


@pytest.mark.parametrize(
    'sections,expected',
    [
        (set(), set()),
        ({'a_b'}, {'A b'}),
        ({'a-b'}, {'A b'}),
        ({'c modelling'}, {'C modeling'}),
        ({'a_b', 'c modelling'}, {'A b', 'C modeling'}),
    ],
)
def test_builder_to_document_sections(sections: Set[str], expected: Set[str]):
  mock_builder = mock.MagicMock()
  instance = document_datasets.BuilderToDocument(
      sections=sections,
      namespace=None,
      builder=mock_builder,
      config_builders=[],
  )
  assert instance.sections == expected



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DatasetInfo records the information we know about a dataset.

This includes things that we know about the dataset statically, i.e.:
 - schema
 - description
 - canonical location
 - does it have validation and tests splits
 - size
 - etc.

This also includes the things that can and should be computed once we've
processed the dataset as well:
 - number of examples (in each split)
 - feature statistics (in each split)
 - etc.
"""

from __future__ import annotations

import abc
import dataclasses
import json
import os
import posixpath
import tempfile
import time
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from absl import logging
from etils import epath
from tensorflow_datasets.core import constants
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.features import feature as feature_lib
from tensorflow_datasets.core.features import top_level_feature
from tensorflow_datasets.core.proto import dataset_info_pb2
from tensorflow_datasets.core.utils import file_utils
from tensorflow_datasets.core.utils import gcs_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

from google.protobuf import json_format

# TODO(b/109648354): Remove the "pytype: disable" comment.
Nest = Union[Tuple["Nest", ...], Dict[str, "Nest"], str]  # pytype: disable=not-supported-yet
SupervisedKeysType = Union[Tuple[Nest, Nest], Tuple[Nest, Nest, Nest]]


def dataset_info_path(dataset_info_dir: epath.PathLike) -> epath.Path:
  return epath.Path(dataset_info_dir) / constants.DATASET_INFO_FILENAME


def license_path(dataset_info_dir: epath.PathLike) -> epath.Path:
  return epath.Path(dataset_info_dir) / constants.LICENSE_FILENAME


class Metadata(dict, metaclass=abc.ABCMeta):
  """Abstract base class for DatasetInfo metadata container.

  `builder.info.metadata` allows the dataset to expose additional general
  information about the dataset which are not specific to a feature or
  individual example.

  To implement the interface, overwrite `save_metadata` and
  `load_metadata`.

  See `tfds.core.MetadataDict` for a simple implementation that acts as a
  dict that saves data to/from a JSON file.
  """

  @abc.abstractmethod
  def save_metadata(self, data_dir):
    """Save the metadata."""
    raise NotImplementedError()

  @abc.abstractmethod
  def load_metadata(self, data_dir):
    """Restore the metadata."""
    raise NotImplementedError()


@dataclasses.dataclass()
class DatasetIdentity:
  """Identity of a dataset that completely identifies a dataset."""

  name: str
  version: utils.Version
  data_dir: str
  module_name: str
  config_name: Optional[str] = None
  config_description: Optional[str] = None
  config_tags: Optional[List[str]] = None
  release_notes: Optional[Dict[str, str]] = None

  @classmethod
  def from_builder(cls, builder) -> "DatasetIdentity":
    if builder.builder_config:
      config_name = builder.builder_config.name
      config_description = builder.builder_config.description
      config_tags = builder.builder_config.tags
    else:
      config_name = None
      config_description = None
      config_tags = None
    return cls(
        name=builder.name,
        version=utils.Version(builder.version),
        data_dir=builder.data_dir,
        module_name=str(builder.__module__),
        config_name=config_name,
        config_description=config_description,
        config_tags=config_tags,
        release_notes=builder.release_notes,
    )

  @classmethod
  def from_proto(
      cls,
      info_proto: dataset_info_pb2.DatasetInfo,
      data_dir: str,
  ) -> "DatasetIdentity":
    return cls(
        name=info_proto.name,
        version=utils.Version(info_proto.version),
        data_dir=data_dir,
        module_name=info_proto.module_name,
        config_name=info_proto.config_name,
        config_description=info_proto.config_description,
        config_tags=info_proto.config_tags or [],
        release_notes={k: v for k, v in info_proto.release_notes.items()},
    )


class DatasetInfo(object):
  """Information about a dataset.

  `DatasetInfo` documents datasets, including its name, version, and features.
  See the constructor arguments and properties for a full list.

  Note: Not all fields are known if the dataset hasn't been generated yet
  (before the first `builder.download_and_prepare()` call). For example splits
  names or number of examples might be missing (as they are computed
  at dataset creation time).
  """

  def __init__(
      self,
      *,
      builder: Union[DatasetIdentity, Any],
      description: Optional[str] = None,
      features: Optional[feature_lib.FeatureConnector] = None,
      supervised_keys: Optional[SupervisedKeysType] = None,
      disable_shuffling: bool = False,
      homepage: Optional[str] = None,
      citation: Optional[str] = None,
      metadata: Optional[Metadata] = None,
      license: Optional[str] = None,  # pylint: disable=redefined-builtin
      redistribution_info: Optional[Dict[str, str]] = None,
      split_dict: Optional[splits_lib.SplitDict] = None,
  ):
    # pyformat: disable
    """Constructs DatasetInfo.

    Args:
      builder: `DatasetBuilder` or `DatasetIdentity`. The dataset builder or
        identity will be used to populate this info.
      description: `str`, description of this dataset.
      features: `tfds.features.FeaturesDict`, Information on the feature dict of
        the `tf.data.Dataset()` object from the `builder.as_dataset()` method.
      supervised_keys: Specifies the input structure for supervised learning, if
        applicable for the dataset, used with "as_supervised". The keys
        correspond to the feature names to select in `info.features`. When
        calling `tfds.core.DatasetBuilder.as_dataset()` with
        `as_supervised=True`, the `tf.data.Dataset` object will yield the
        structure defined by the keys passed here, instead of that defined by
        the `features` argument. Typically this is a `(input_key, target_key)`
        tuple, and the dataset yields a tuple of tensors `(input, target)`
        tensors.

        To yield a more complex structure, pass a tuple of `tf.nest` compatible
        structures of feature keys. The resulting `Dataset` will yield
        structures with each key replaced by the coresponding tensor. For
        example, passing a triple of keys would return a dataset
        that yields `(feature, target, sample_weights)` triples for keras.
        Using `supervised_keys=({'a':'a','b':'b'}, 'c')` would create a dataset
        yielding a tuple with a dictionary of features in the `features`
        position.

        Note that selecting features in nested `tfds.features.FeaturesDict`
        objects is not supported.
      disable_shuffling: `bool`, specify whether to shuffle the examples.
      homepage: `str`, optional, the homepage for this dataset.
      citation: `str`, optional, the citation to use for this dataset.
      metadata: `tfds.core.Metadata`, additonal object which will be
        stored/restored with the dataset. This allows for storing additional
        information with the dataset.
      license: license of the dataset.
      redistribution_info: information needed for redistribution, as specified
        in `dataset_info_pb2.RedistributionInfo`. The content of the `license`
        subfield will automatically be written to a LICENSE file stored with the
        dataset.
      split_dict: information about the splits in this dataset.
    """
    # pyformat: enable
    self._builder_or_identity = builder
    if isinstance(builder, DatasetIdentity):
      self._identity = builder
    else:
      self._identity = DatasetIdentity.from_builder(builder)

    if redistribution_info is not None:
      redistribution_info = dataset_info_pb2.RedistributionInfo(
          license=utils.dedent(license or redistribution_info.pop("license")),
          **redistribution_info,
      )

    self._info_proto = dataset_info_pb2.DatasetInfo(
        name=self._identity.name,
        description=utils.dedent(description),
        version=str(self._identity.version),
        release_notes=self._identity.release_notes,
        disable_shuffling=disable_shuffling,
        config_name=self._identity.config_name,
        config_description=self._identity.config_description,
        config_tags=self._identity.config_tags,
        citation=utils.dedent(citation),
        module_name=self._identity.module_name,
        redistribution_info=redistribution_info,
    )

    if homepage:
      self._info_proto.location.urls[:] = [homepage]

    if features:
      if not isinstance(features, top_level_feature.TopLevelFeature):
        raise ValueError(
            "DatasetInfo.features only supports FeaturesDict or Sequence at "
            "the top-level. Got {}".format(features)
        )
    self._features = features
    self._splits = splits_lib.SplitDict([])
    if split_dict:
      self.set_splits(split_dict)
    if supervised_keys is not None:
      self._info_proto.supervised_keys.CopyFrom(
          _supervised_keys_to_proto(supervised_keys)
      )

    if metadata and not isinstance(metadata, Metadata):
      raise ValueError(
          "Metadata should be a `tfds.core.Metadata` instance. Received "
          "{}".format(metadata)
      )
    self._metadata = metadata

    # Is this object initialized with both the static and the dynamic data?
    self._fully_initialized = False

  @property
  def _builder(self) -> Any:
    logging.warning("DEPRECATED: please do not use _builder as this may change")
    return self._builder_or_identity

  @classmethod
  def from_proto(
      cls,
      builder,
      proto: dataset_info_pb2.DatasetInfo,
  ) -> "DatasetInfo":
    """Instantiates DatasetInfo from the given builder and proto."""
    if builder.builder_config:
      assert builder.builder_config.name == proto.config_name
    assert str(builder.version) == proto.version
    features = None
    if proto.HasField("features"):
      features = feature_lib.FeatureConnector.from_proto(proto.features)
    supervised_keys = None
    if proto.HasField("supervised_keys"):
      supervised_keys = _supervised_keys_from_proto(proto.supervised_keys)
    filename_template = naming.ShardedFileTemplate(
        dataset_name=builder.name,
        data_dir=builder.data_dir,
        filetype_suffix=proto.file_format or "tfrecord",
    )
    return cls(
        builder=builder,
        description=proto.description,
        features=features,
        supervised_keys=supervised_keys,
        disable_shuffling=proto.disable_shuffling,
        citation=proto.citation,
        license=proto.redistribution_info.license,
        split_dict=splits_lib.SplitDict.from_proto(
            repeated_split_infos=proto.splits,
            filename_template=filename_template,
        ),
    )

  @property
  def as_proto(self) -> dataset_info_pb2.DatasetInfo:
    return self._info_proto

  @property
  def as_proto_with_features(self) -> dataset_info_pb2.DatasetInfo:
    info_proto = dataset_info_pb2.DatasetInfo()
    info_proto.CopyFrom(self._info_proto)
    info_proto.features.CopyFrom(self.features.to_proto())
    return info_proto

  @property
  def name(self) -> str:
    return self._identity.name

  @property
  def config_name(self) -> str:
    return self._info_proto.config_name

  @property
  def config_description(self) -> str | None:
    return self._identity.config_description

  @property
  def config_tags(self) -> List[str] | None:
    return self._identity.config_tags

  @property
  def full_name(self):
    """Full canonical name: (<dataset_name>/<config_name>/<version>)."""
    names = [self.name]
    if self.config_name:
      names.append(self.config_name)
    names.append(str(self.version))
    return posixpath.join(*names)

  @property
  def description(self):
    return self.as_proto.description

  @property
  def version(self):
    return self._identity.version

  @property
  def release_notes(self) -> Optional[Dict[str, str]]:
    return self._identity.release_notes

  @property
  def disable_shuffling(self) -> bool:
    return self.as_proto.disable_shuffling

  @property
  def homepage(self):
    urls = self.as_proto.location.urls
    tfds_homepage = f"https://www.tensorflow.org/datasets/catalog/{self.name}"
    return urls and urls[0] or tfds_homepage

  @property
  def citation(self) -> str:
    return self.as_proto.citation

  @property
  def data_dir(self):
    return self._identity.data_dir

  @property
  def dataset_size(self) -> utils.Size:
    """Generated dataset files size, in bytes."""
    # For old datasets, maybe empty.
    return utils.Size(sum(split.num_bytes for split in self.splits.values()))

  @property
  def download_size(self) -> utils.Size:
    """Downloaded files size, in bytes."""
    # Fallback to deprecated `size_in_bytes` if `download_size` is empty.
    return utils.Size(
        self.as_proto.download_size or self.as_proto.size_in_bytes
    )

  @download_size.setter
  def download_size(self, size):
    self.as_proto.download_size = size

  @property
  def features(self):
    return self._features

  @property
  def metadata(self) -> Optional[Metadata]:
    return self._metadata

  @property
  def supervised_keys(self) -> Optional[SupervisedKeysType]:
    if not self.as_proto.HasField("supervised_keys"):
      return None
    supervised_keys = self.as_proto.supervised_keys
    return _supervised_keys_from_proto(supervised_keys)

  @property
  def redistribution_info(self):
    return self.as_proto.redistribution_info

  @property
  def module_name(self) -> str:
    return self._identity.module_name

  @property
  def file_format(self) -> Optional[file_adapters.FileFormat]:
    if not self.as_proto.file_format:
      return None
    return file_adapters.FileFormat(self.as_proto.file_format)

  def set_file_format(
      self,
      file_format: Union[None, str, file_adapters.FileFormat],
      override: bool = False,
  ) -> None:
    """Internal function to define the file format.

    The file format is set during `FileReaderBuilder.__init__`,
    not `DatasetInfo.__init__`.

    Args:
      file_format: The file format.
      override: Whether the file format should be overridden if it is already
        set.

    Raises:
      ValueError: if the file format was already set and the `override`
      parameter was False.
      RuntimeError: if an incorrect combination of options is given, e.g.
      `override=True` when the DatasetInfo is already fully initialized.
    """
    # If file format isn't present already, fallback to `DEFAULT_FILE_FORMAT`
    file_format = (
        file_format  # Format explicitly given: tfds.builder(..., file_format=x)
        or self.file_format  # Format restored from dataset_info.json
        or file_adapters.DEFAULT_FILE_FORMAT
    )
    file_format = file_adapters.FileFormat.from_value(file_format)

    # If the file format has been set once, file format should be consistent
    if not override and self.file_format and self.file_format != file_format:
      raise ValueError(
          f"File format is already set to {self.file_format}. Got {file_format}"
      )
    if override and self._fully_initialized:
      raise RuntimeError(
          "Cannot override the file format "
          "when the DatasetInfo is already fully initialized!"
      )
    self._info_proto.file_format = file_format.value

  @property
  def splits(self) -> splits_lib.SplitDict:
    return self._splits

  def set_splits(self, split_dict: splits_lib.SplitDict) -> None:
    """Split setter (private method)."""
    for split, split_info in split_dict.items():
      if isinstance(split_info, splits_lib.MultiSplitInfo):
        # When splits are from multiple folders, the dataset can be different.
        continue
      if (
          split_info.filename_template
          and self.name != split_info.filename_template.dataset_name
      ):
        raise AssertionError(
            f"SplitDict contains SplitInfo for split {split} whose "
            "dataset_name does not match to the dataset name in dataset_info. "
            f"{self.name} != {split_info.filename_template.dataset_name}"
        )

    # If the statistics have been pre-loaded, forward the statistics
    # into the new split_dict. Also add the filename template if it's not set.
    new_split_infos = []
    incomplete_filename_template = naming.ShardedFileTemplate(
        dataset_name=self.name,
        data_dir=self.data_dir,
        filetype_suffix=(
            self.as_proto.file_format or file_adapters.DEFAULT_FILE_FORMAT.value
        ),
    )
    for split_info in split_dict.values():
      if isinstance(split_info, splits_lib.MultiSplitInfo):
        new_split_infos.append(split_info)
        continue
      old_split_info = self._splits.get(split_info.name)
      if (
          not split_info.statistics.ByteSize()
          and old_split_info
          and old_split_info.statistics.ByteSize()
          and old_split_info.shard_lengths == split_info.shard_lengths
      ):
        split_info = split_info.replace(statistics=old_split_info.statistics)
      if not split_info.filename_template:
        filename_template = incomplete_filename_template.replace(
            split=split_info.name
        )
        split_info = split_info.replace(filename_template=filename_template)
      new_split_infos.append(split_info)

    # Update the dictionary representation.
    self._splits = splits_lib.SplitDict(new_split_infos)

    # Update the proto
    # Note that the proto should not be saved or used for multi-folder datasets.
    del self.as_proto.splits[:]  # Clear previous
    for split_info in self._splits.values():
      if isinstance(split_info, splits_lib.MultiSplitInfo):
        for si in split_info.split_infos:
          self.as_proto.splits.add().CopyFrom(si.to_proto())
      else:
        self.as_proto.splits.add().CopyFrom(split_info.to_proto())

  def update_data_dir(self, data_dir: str) -> None:
    """Updates the data dir for each split."""
    new_split_infos = []
    for split_info in self._splits.values():
      if isinstance(split_info, splits_lib.MultiSplitInfo):
        raise RuntimeError(
            "Updating the data_dir for MultiSplitInfo is not supported!"
        )
      if not split_info.filename_template:
        continue
      filename_template = split_info.filename_template.replace(
          data_dir=data_dir
      )
      new_split_info = split_info.replace(filename_template=filename_template)
      new_split_infos.append(new_split_info)
    self.set_splits(splits_lib.SplitDict(new_split_infos))

  @property
  def initialized(self) -> bool:
    """Whether DatasetInfo has been fully initialized."""
    return self._fully_initialized

  @property
  def as_json(self) -> str:
    return json_format.MessageToJson(self.as_proto, sort_keys=True)

  def write_to_directory(
      self, dataset_info_dir: epath.PathLike, all_metadata=True
  ) -> None:
    """Write `DatasetInfo` as JSON to `dataset_info_dir` + labels & features.

    Args:
      dataset_info_dir: path to directory in which to save the
        `dataset_info.json` file, as well as `features.json` and `*.labels.txt`
        if applicable.
      all_metadata: defaults to True. If False, will not write metadata which
        may have an impact on how the data is read (features.json). Should be
        set to True whenever `write_to_directory` is called for the first time
        for a new dataset.
    """
    if all_metadata:
      # Save the features structure & metadata (vocabulary, labels,...)
      if self.features:
        self.features.save_config(dataset_info_dir)

    # Save any additional metadata
    if self.metadata is not None:
      self.metadata.save_metadata(dataset_info_dir)

    if self.redistribution_info.license:
      license_path(dataset_info_dir).write_text(
          self.redistribution_info.license
      )

    dataset_info_path(dataset_info_dir).write_text(self.as_json)

  def read_from_directory(self, dataset_info_dir: epath.PathLike) -> None:
    """Update DatasetInfo from the metadata files in `dataset_info_dir`.

    This function updates all the dynamically generated fields (num_examples,
    hash, time of creation,...) of the DatasetInfo.

    This will overwrite all previous metadata.

    Args:
      dataset_info_dir: The directory containing the metadata file. This should
        be the root directory of a specific dataset version.

    Raises:
      FileNotFoundError: If the dataset_info.json can't be found.
    """
    logging.info("Load dataset info from %s", dataset_info_dir)

    # Load the metadata from disk
    try:
      parsed_proto = read_from_json(dataset_info_path(dataset_info_dir))
    except Exception as e:
      raise FileNotFoundError(
          "Tried to load `DatasetInfo` from a directory which does not exist or"
          " does not contain `dataset_info.json`. Please delete the directory "
          f"`{dataset_info_dir}`  if you are trying to re-generate the "
          "dataset."
      ) from e

    if str(self.version) != parsed_proto.version:
      raise AssertionError(
          "The constructed DatasetInfo instance and the restored proto version "
          "do not match. Builder version: {}. Proto version: {}".format(
              self.version, parsed_proto.version
          )
      )

    self._identity = DatasetIdentity.from_proto(
        info_proto=parsed_proto, data_dir=dataset_info_dir
    )

    # Update splits
    filename_template = naming.ShardedFileTemplate(
        dataset_name=self.name,
        data_dir=self.data_dir,
        filetype_suffix=parsed_proto.file_format or "tfrecord",
    )
    split_dict = splits_lib.SplitDict.from_proto(
        repeated_split_infos=parsed_proto.splits,
        filename_template=filename_template,
    )
    self.set_splits(split_dict)

    # Restore the feature metadata (vocabulary, labels names,...)
    if self.features:
      self.features.load_metadata(dataset_info_dir)
    # For `ReadOnlyBuilder`, reconstruct the features from the config.
    elif feature_lib.make_config_path(dataset_info_dir).exists():
      self._features = feature_lib.FeatureConnector.from_config(
          dataset_info_dir
      )
    # Restore the MetaDataDict from metadata.json if there is any
    if (
        self.metadata is not None
        or _metadata_filepath(dataset_info_dir).exists()
    ):
      # If the dataset was loaded from file, self.metadata will be `None`, so
      # we create a MetadataDict first.
      if self.metadata is None:
        self._metadata = MetadataDict()
      self.metadata.load_metadata(dataset_info_dir)

    # Update fields which are not defined in the code. This means that
    # the code will overwrite fields which are present in
    # dataset_info.json.
    fields_taken_from_code = []
    for field_name, field in self.as_proto.DESCRIPTOR.fields_by_name.items():
      field_value = getattr(self._info_proto, field_name)
      field_value_restored = getattr(parsed_proto, field_name)

      try:
        is_defined = self._info_proto.HasField(field_name)
      except ValueError:
        is_defined = bool(field_value)

      try:
        is_defined_in_restored = parsed_proto.HasField(field_name)
      except ValueError:
        is_defined_in_restored = bool(field_value_restored)

      # If field is defined in code, we ignore the value.
      if is_defined:
        if field_value != field_value_restored:
          fields_taken_from_code.append(field_name)
        continue
      # If the field is also not defined in JSON file, we do nothing
      if not is_defined_in_restored:
        continue
      # Otherwise, we restore the dataset_info.json value
      if field.type == field.TYPE_MESSAGE:
        field_value.MergeFrom(field_value_restored)
      elif field.label == field.LABEL_REPEATED:
        del field_value[:]
        field_value.extend(field_value_restored)
      else:
        setattr(self._info_proto, field_name, field_value_restored)

    if fields_taken_from_code:
      logging.info(
          (
              "For '%s': fields info.[%s] differ on disk and in the code. "
              "Keeping the one from code."
          ),
          self.full_name,
          ", ".join(fields_taken_from_code),
      )

    # Mark as fully initialized.
    self._fully_initialized = True

  def add_file_data_source_access(
      self,
      path: Union[epath.PathLike, Iterable[epath.PathLike]],
      url: Optional[str] = None,
  ) -> None:
    """Records that the given query was used to generate this dataset.

    Arguments:
      path: path or paths of files that were read. Can be a file pattern.
        Multiple paths or patterns can be specified as a comma-separated string
        or a list.
      url:  URL referring to the data being used.
    """
    access_timestamp_ms = _now_in_milliseconds()
    if isinstance(path, str) or isinstance(path, epath.Path):
      path = os.fspath(path).split(",")
    for p in path:
      for file in file_utils.expand_glob(p):
        self._info_proto.data_source_accesses.append(
            dataset_info_pb2.DataSourceAccess(
                access_timestamp_ms=access_timestamp_ms,
                file_system=dataset_info_pb2.FileSystem(path=os.fspath(file)),
                url=dataset_info_pb2.Url(url=url),
            )
        )

  def add_url_access(
      self,
      url: str,
      checksum: Optional[str] = None,
  ) -> None:
    """Records the URL used to generate this dataset."""
    self._info_proto.data_source_accesses.append(
        dataset_info_pb2.DataSourceAccess(
            access_timestamp_ms=_now_in_milliseconds(),
            url=dataset_info_pb2.Url(url=url, checksum=checksum),
        )
    )

  def add_sql_data_source_access(
      self,
      sql_query: str,
  ) -> None:
    """Records that the given query was used to generate this dataset."""
    self._info_proto.data_source_accesses.append(
        dataset_info_pb2.DataSourceAccess(
            access_timestamp_ms=_now_in_milliseconds(),
            sql_query=dataset_info_pb2.SqlQuery(sql_query=sql_query),
        )
    )

  def add_tfds_data_source_access(
      self,
      dataset_reference: naming.DatasetReference,
      url: Optional[str] = None,
  ) -> None:
    """Records that the given query was used to generate this dataset.

    Args:
      dataset_reference:
      url: a URL referring to the TFDS dataset.
    """
    self._info_proto.data_source_accesses.append(
        dataset_info_pb2.DataSourceAccess(
            access_timestamp_ms=_now_in_milliseconds(),
            tfds_dataset=dataset_info_pb2.TfdsDatasetReference(
                name=dataset_reference.dataset_name,
                config=dataset_reference.config,
                version=str(dataset_reference.version),
                data_dir=dataset_reference.data_dir,
                ds_namespace=dataset_reference.namespace,
            ),
            url=dataset_info_pb2.Url(url=url),
        )
    )

  def initialize_from_bucket(self) -> None:
    """Initialize DatasetInfo from GCS bucket info files."""
    # In order to support Colab, we use the HTTP GCS API to access the metadata
    # files. They are copied locally and then loaded.
    tmp_dir = epath.Path(tempfile.mkdtemp("tfds"))
    data_files = gcs_utils.gcs_dataset_info_files(self.full_name)
    if not data_files:
      return
    logging.info(
        (
            "Load pre-computed DatasetInfo (eg: splits, num examples,...) "
            "from GCS: %s"
        ),
        self.full_name,
    )
    for path in data_files:
      out_fname = tmp_dir / path.name
      epath.Path(path).copy(out_fname)
    self.read_from_directory(tmp_dir)

  def __repr__(self):
    SKIP = object()  # pylint: disable=invalid-name

    splits = _indent(
        "\n".join(
            ["{"]
            + [
                f"    '{k}': {split},"
                for k, split in sorted(self.splits.items())
            ]
            + ["}"]
        )
    )

    if self._info_proto.config_description:
      config_description = _indent(
          f'"""\n{self._info_proto.config_description}\n"""'
      )
    else:
      config_description = SKIP

    if self._info_proto.config_tags:
      config_tags = ", ".join(self.config_tags)
    else:
      config_tags = SKIP

    file_format_str = (
        self.file_format.value
        if self.file_format
        else file_adapters.DEFAULT_FILE_FORMAT.value
    )
    lines = ["tfds.core.DatasetInfo("]
    for key, value in [
        ("name", repr(self.name)),
        ("full_name", repr(self.full_name)),
        ("description", _indent(f'"""\n{self.description}\n"""')),
        ("config_description", config_description),
        ("config_tags", config_tags),
        ("homepage", repr(self.homepage)),
        ("data_dir", repr(self.data_dir)),
        ("file_format", file_format_str),
        ("download_size", self.download_size),
        ("dataset_size", self.dataset_size),
        ("features", _indent(repr(self.features))),
        ("supervised_keys", self.supervised_keys),
        ("disable_shuffling", self.disable_shuffling),
        ("splits", splits),
        ("citation", _indent(f'"""{self.citation}"""')),
        # Proto add a \n that we strip.
        ("redistribution_info", str(self.redistribution_info).strip() or SKIP),
    ]:
      if value != SKIP:
        lines.append(f"    {key}={value},")
    lines.append(")")
    return "\n".join(lines)


def _nest_to_proto(nest: Nest) -> dataset_info_pb2.SupervisedKeys.Nest:
  """Creates a `SupervisedKeys.Nest` from a limited `tf.nest` style structure.

  Args:
    nest: A `tf.nest` structure of tuples, dictionaries or string feature keys.

  Returns:
    The same structure as a `SupervisedKeys.Nest` proto.
  """
  nest_type = type(nest)
  proto = dataset_info_pb2.SupervisedKeys.Nest()
  if nest_type is tuple:
    for item in nest:
      proto.tuple.items.append(_nest_to_proto(item))
  elif nest_type is dict:
    nest = {key: _nest_to_proto(value) for key, value in nest.items()}
    proto.dict.CopyFrom(dataset_info_pb2.SupervisedKeys.Dict(dict=nest))
  elif nest_type is str:
    proto.feature_key = nest
  else:
    raise ValueError(
        "The nested structures in `supervised_keys` must only "
        "contain instances of (tuple, dict, str), no subclasses.\n"
        f"Found type: {nest_type}"
    )

  return proto


def _supervised_keys_to_proto(
    keys: SupervisedKeysType,
) -> dataset_info_pb2.SupervisedKeys:
  """Converts a `supervised_keys` tuple to a SupervisedKeys proto."""
  if not isinstance(keys, tuple) or len(keys) not in [2, 3]:
    raise ValueError(
        "`supervised_keys` must contain a tuple of 2 or 3 elements.\n"
        f"got: {keys!r}"
    )

  proto = dataset_info_pb2.SupervisedKeys(
      tuple=dataset_info_pb2.SupervisedKeys.Tuple(
          items=(_nest_to_proto(key) for key in keys)
      )
  )
  return proto


def _nest_from_proto(proto: dataset_info_pb2.SupervisedKeys.Nest) -> Nest:
  """Creates a `tf.nest` style structure from a `SupervisedKeys.Nest` proto.

  Args:
    proto: A `SupervisedKeys.Nest` proto.

  Returns:
    The proto converted to a `tf.nest` style structure of tuples, dictionaries
    or strings.
  """
  if proto.HasField("tuple"):
    return tuple(_nest_from_proto(item) for item in proto.tuple.items)
  elif proto.HasField("dict"):
    return {
        key: _nest_from_proto(value)
        for key, value in sorted(proto.dict.dict.items())
    }
  elif proto.HasField("feature_key"):
    return proto.feature_key
  else:
    raise ValueError(
        "`SupervisedKeys.Nest` proto must contain one of "
        f"(tuple, dict, feature_key). Got: {proto}"
    )


def _supervised_keys_from_proto(
    proto: dataset_info_pb2.SupervisedKeys,
) -> SupervisedKeysType:
  """Converts a `SupervisedKeys` proto back to a simple python tuple."""
  if proto.input and proto.output:
    return (proto.input, proto.output)
  elif proto.tuple:
    return tuple(_nest_from_proto(item) for item in proto.tuple.items)
  else:
    raise ValueError(
        "A `SupervisedKeys` proto must have either `input` and "
        "`output` defined, or `tuple`, got: {proto}"
    )


def _indent(content):
  """Add indentation to all lines except the first."""
  lines = content.split("\n")
  return "\n".join([lines[0]] + ["    " + l for l in lines[1:]])


def _populate_shape(shape_or_dict, prefix, schema_features):
  """Populates shape in the schema."""
  if isinstance(shape_or_dict, (tuple, list)):
    feature_name = "/".join(prefix)
    if shape_or_dict and feature_name in schema_features:
      schema_feature = schema_features[feature_name]
      schema_feature.ClearField("shape")
      for dim in shape_or_dict:
        # We denote `None`s as -1 in the shape proto.
        schema_feature.shape.dim.add().size = -1 if dim is None else dim
    return
  for name, val in shape_or_dict.items():
    prefix.append(name)
    _populate_shape(val, prefix, schema_features)
    prefix.pop()


def get_dataset_feature_statistics(builder, split):
  """Calculate statistics for the specified split."""
  tfdv = lazy_imports_lib.lazy_imports.tensorflow_data_validation
  # TODO(epot): Avoid hardcoding file format.
  filetype_suffix = "tfrecord"
  if filetype_suffix not in ["tfrecord", "csv"]:
    raise ValueError(
        "Cannot generate statistics for filetype {}".format(filetype_suffix)
    )
  filename_template = naming.ShardedFileTemplate(
      data_dir=builder.data_dir,
      dataset_name=builder.name,
      split=split,
      filetype_suffix=filetype_suffix,
  )
  filepattern = filename_template.sharded_filepaths_pattern()
  # Avoid generating a large number of buckets in rank histogram
  # (default is 1000).
  stats_options = tfdv.StatsOptions(
      num_top_values=10,
      num_rank_histogram_buckets=10,
      use_sketch_based_topk_uniques=False,
  )
  if filetype_suffix == "csv":
    statistics = tfdv.generate_statistics_from_csv(
        filepattern, stats_options=stats_options
    )
  else:
    statistics = tfdv.generate_statistics_from_tfrecord(
        filepattern, stats_options=stats_options
    )
  schema = tfdv.infer_schema(statistics)
  schema_features = {feature.name: feature for feature in schema.feature}
  # Override shape in the schema.
  for feature_name, feature in builder.info.features.items():
    _populate_shape(feature.shape, [feature_name], schema_features)

  # Remove legacy field.
  if getattr(schema, "generate_legacy_feature_spec", None) is not None:
    schema.ClearField("generate_legacy_feature_spec")
  return statistics.datasets[0], schema


def read_from_json(path: epath.PathLike) -> dataset_info_pb2.DatasetInfo:
  """Read JSON-formatted proto into DatasetInfo proto.

  Args:
    path: the path of the dataset info json file.

  Returns:
    the DatasetInfo proto.

  Raises:
    FileNotFoundError: If the builder_dir does not exist.
  """
  try:
    json_str = epath.Path(path).read_text()
  except OSError as e:
    raise FileNotFoundError(f"Could not load dataset info from {path}") from e
  # Parse it back into a proto.
  parsed_proto = json_format.Parse(json_str, dataset_info_pb2.DatasetInfo())
  return parsed_proto


def read_proto_from_builder_dir(
    builder_dir: epath.PathLike,
) -> dataset_info_pb2.DatasetInfo:
  """Reads the dataset info from the given builder dir.

  Args:
    builder_dir: The folder that contains the dataset info files.

  Returns:
    The DatasetInfo proto as read from the builder dir.

  Raises:
    FileNotFoundError: If the builder_dir does not exist.
  """
  builder_dir = epath.Path(builder_dir).expanduser()
  info_path = builder_dir / constants.DATASET_INFO_FILENAME
  return read_from_json(info_path)


def read_full_proto_from_builder_dir(
    builder_dir: epath.PathLike,
) -> dataset_info_pb2.DatasetInfo:
  """Reads the dataset info from the given builder dir.

  Also reads data from the features files and puts it inside the dataset info.

  Args:
    builder_dir: The folder that contains the dataset info files.

  Returns:
    The DatasetInfo proto including feature information as read from the builder
    dir.

  Raises:
    FileNotFoundError: If the builder_dir does not exist or it doesn't contain
    dataset_info.json.
  """
  builder_dir = epath.Path(builder_dir).expanduser()
  info_path = builder_dir / constants.DATASET_INFO_FILENAME
  info_proto = read_from_json(info_path)
  if not info_proto.HasField("features"):
    try:
      features = feature_lib.FeatureConnector.from_config(
          os.fspath(builder_dir)
      )
      info_proto.features.CopyFrom(features.to_proto())
    except IOError:
      logging.warning("Could not read features from %s", builder_dir)
  return info_proto


def pack_as_supervised_ds(
    ds: tf.data.Dataset,
    ds_info: DatasetInfo,
) -> tf.data.Dataset:
  """Pack `(input, label)` dataset as `{'key0': input, 'key1': label}`."""
  if (
      ds_info.supervised_keys
      and isinstance(ds.element_spec, tuple)
      and len(ds.element_spec) == 2
  ):
    x_key, y_key = ds_info.supervised_keys
    ds = ds.map(lambda x, y: {x_key: x, y_key: y})
    return ds
  else:  # If dataset isn't a supervised tuple (input, label), return as-is
    return ds


def _metadata_filepath(data_dir: epath.PathLike) -> epath.Path:
  return epath.Path(data_dir) / constants.METADATA_FILENAME


def _now_in_milliseconds() -> int:
  return time.time_ns() // 1000000


class MetadataDict(Metadata, dict):
  """A `tfds.core.Metadata` object that acts as a `dict`.

  By default, the metadata will be serialized as JSON.
  """

  def save_metadata(self, data_dir):
    """Save the metadata."""
    with _metadata_filepath(data_dir).open(mode="w") as f:
      json.dump(self, f)

  def load_metadata(self, data_dir):
    """Restore the metadata."""
    self.clear()
    with _metadata_filepath(data_dir).open(mode="r") as f:
      self.update(json.load(f))


class BeamMetadataDict(MetadataDict):
  """A `tfds.core.Metadata` object supporting Beam-generated datasets."""

  def __init__(self, *args, **kwargs):
    super(BeamMetadataDict, self).__init__(*args, **kwargs)
    self._tempdir = epath.Path(tempfile.mkdtemp("tfds_beam_metadata"))

  def _temp_filepath(self, key) -> epath.Path:
    return self._tempdir / f"{key}.json"

  def __setitem__(self, key, item):
    """Creates write sink for beam PValues or sets value of key in `dict`.

    If the item is a PValue, it is expected to contain exactly one element,
    which will be written out as a temporary JSON file once the beam pipeline
    runs. These outputs will be loaded and stored in a single JSON when
    `save_metadata` is called after the pipeline completes.

    Args:
      key: hashable type, the key for the item.
      item: `beam.pvalue.PValue` or other, the metadata value.
    """
    beam = lazy_imports_lib.lazy_imports.apache_beam
    if isinstance(item, beam.PTransform):
      # Implementing Beam support might be possible but would
      # require very careful implementation to avoid computing the
      # PTransform twice (once for the split and once for the metadata).
      raise NotImplementedError(
          "`tfds.core.BeamMetadataDict` can't be used on `beam.PTransform`, "
          "only on `beam.PCollection`. See `_generate_examples` doc on how "
          "to use `beam.PCollection`, or wrap your `_generate_examples` inside "
          f"a @beam.ptransform_fn. Got: {key}: {item}"
      )
    elif isinstance(item, beam.pvalue.PValue):
      if key in self:
        raise ValueError("Already added PValue with key: %s" % key)
      logging.info("Lazily adding metadata item with Beam: %s", key)

      def _to_json(item_list):
        if len(item_list) != 1:
          raise ValueError(
              "Each metadata PValue must contain a single element. Got %d."
              % len(item_list)
          )
        item = item_list[0]
        return json.dumps(item)

      _ = (
          item
          | "metadata_%s_tolist" % key >> beam.combiners.ToList()
          | "metadata_%s_tojson" % key >> beam.Map(_to_json)
          | "metadata_%s_write" % key
          >> beam.io.WriteToText(
              os.fspath(self._temp_filepath(key)),
              num_shards=1,
              shard_name_template="",
          )
      )
    super(BeamMetadataDict, self).__setitem__(key, item)

  def save_metadata(self, data_dir):
    """Save the metadata inside the beam job."""
    beam = lazy_imports_lib.lazy_imports.apache_beam
    for key, item in self.items():
      if isinstance(item, beam.pvalue.PValue):
        with self._temp_filepath(key).open(mode="r") as f:
          self[key] = json.load(f)
    self._tempdir.rmtree()
    super(BeamMetadataDict, self).save_metadata(data_dir)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Meta register that uses other registers."""

import dataclasses
import difflib
import functools
import os
from typing import Any, Iterable, List, Mapping, Type

from absl import logging
from etils import epath
from tensorflow_datasets.core import dataset_builder
# Make sure that github paths are registered
from tensorflow_datasets.core import github_api  # pylint: disable=unused-import
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import registered
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.community import register_base
from tensorflow_datasets.core.community import register_package
from tensorflow_datasets.core.community import register_path
from tensorflow_datasets.core.utils import gcs_utils
import toml


def _as_path_list(path_or_paths: utils.ListOrElem[str]) -> List[epath.Path]:
  if isinstance(path_or_paths, list):
    return [epath.Path(p) for p in path_or_paths]
  else:
    return [epath.Path(path_or_paths)]


def _load_register_for_paths(
    namespace: str,
    paths: List[epath.Path],
) -> List[register_base.BaseRegister]:
  """Returns a list of registers for the given paths."""
  code_paths = []
  data_paths = []
  for path in paths:
    if os.fspath(path).startswith('github'):
      code_paths.append(path)
    else:
      data_paths.append(path)

  if code_paths and data_paths:
    raise RuntimeError(
        f'Both a path containing code ({code_paths}) and '
        f'a path containing data ({data_paths} are specified. '
        'This is not supported'
    )

  registers = []
  if data_paths:
    register = register_path.DataDirRegister(
        namespace_to_data_dirs={namespace: data_paths}
    )
    registers.append(register)
  if code_paths:
    registers.append(
        register_package.PackageRegister(
            path=gcs_utils.GCS_COMMUNITY_INDEX_PATH
        )
    )
  return registers


@dataclasses.dataclass(eq=True, frozen=True)
class NamespaceConfig:
  """Config what namespaces there are and where to find their datasets.

  The config file should be in `.toml` format and have the follow structure:

  ```toml
  [Namespaces]
  huggingface='github://huggingface/datasets/tree/master/datasets'
  kaggle=['/path/to/datasets/', '/path/to/more_datasets/']
  tensorflow_graphics='gs://tensorflow-graphics/datasets'
  ```
  """

  config_path: epath.Path

  @functools.lru_cache()
  def registers_per_namespace(
      self,
  ) -> Mapping[str, List[register_base.BaseRegister]]:
    """Returns the registry containing all repositories in the given config.

    Raises:
      RuntimeError: when the config contains errors.
    """
    config = toml.loads(self.config_path.read_text())
    registers_per_namespace = {}
    for namespace, path_or_paths in config['Namespaces'].items():
      if namespace in registers_per_namespace:
        raise RuntimeError(
            f'Namespace {namespace} is defined twice in config'
            f' {self.config_path}'
        )
      registers_per_namespace[namespace] = _load_register_for_paths(
          namespace=namespace, paths=_as_path_list(path_or_paths)
      )
    return registers_per_namespace


@dataclasses.dataclass()
class DatasetRegistry(register_base.BaseRegister):
  """Registry of dataset registries.

  Each namespace is associated with one or more registers. Those registers can
  be any type of register, e.g. code on Github or datasets stored somewhere.

  Attributes:
    namespace_config: config where to find the datasets of a namespace.
    registers_per_namespace: per namespace a list of registers it consists of.
  """

  namespace_config: NamespaceConfig

  @property
  def registers_per_namespace(
      self,
  ) -> Mapping[str, List[register_base.BaseRegister]]:
    return self.namespace_config.registers_per_namespace()

  def has_namespace(self, namespace: str) -> bool:
    if not namespace:
      return False
    return namespace in self.registers_per_namespace

  def list_namespaces(self) -> List[str]:
    return sorted(self.registers_per_namespace.keys())

  def list_builders(self) -> List[str]:
    builders = []
    for registers in self.registers_per_namespace.values():
      for register in registers:
        builders.extend(register.list_builders())
    return builders

  def list_dataset_references(self) -> Iterable[naming.DatasetReference]:
    for registers in self.registers_per_namespace.values():
      for register in registers:
        try:
          yield from register.list_dataset_references()
        except Exception:  # pylint: disable=broad-except
          logging.exception(
              'Exception while getting dataset references from register %s',
              register,
          )

  def list_builders_per_namespace(self, namespace: str) -> List[str]:
    """Lists the builders available for a specific namespace."""
    builders = []
    if self.has_namespace(namespace):
      for register in self.registers_per_namespace[namespace]:
        builders.extend(register.list_builders())
    return builders

  def _get_list_builders_context(self, name: naming.DatasetName) -> str:
    """Adds relevant information to the error context."""
    # Add list of available datasets to error context.
    all_datasets = self.list_builders_per_namespace(name.namespace)
    all_datasets_str = '\n\t- '.join([''] + all_datasets)
    error_msg = (
        f'Available datasets under the same namespace:{all_datasets_str}\n'
    )
    # Add closest match to error context.
    close_matches = difflib.get_close_matches(str(name), all_datasets, n=1)
    if close_matches:
      error_msg += f'\nDid you mean: {name} -> {close_matches[0]} ?\n'
    return error_msg

  def _get_registers(
      self, name: naming.DatasetName
  ) -> List[register_base.BaseRegister]:
    """Returns all available registers for a given namespace, if any.

    Args:
      name: str, the namespace's name.

    Raises:
      DatasetNotFound error if the namespace is not found.
    """
    if not self.has_namespace(name.namespace):
      error_msg = f'\nNamespace {name.namespace} not found. '
      error_msg += (
          'Note that the namespace should be one of: '
          f'{sorted(self.registers_per_namespace.keys())}.\n'
      )
      close_matches = difflib.get_close_matches(
          name.namespace, self.registers_per_namespace, n=1
      )
      if close_matches:
        error_msg += f'Did you mean: {name.namespace} -> {close_matches[0]} ?\n'
      raise registered.DatasetNotFoundError(error_msg)
    return self.registers_per_namespace[name.namespace]

  def builder_cls(
      self,
      name: naming.DatasetName,
  ) -> Type[dataset_builder.DatasetBuilder]:
    """Loads the builder class for the given dataset.

    Arguments:
      name: the name and namespace of the dataset to load the builder class for.

    Returns:
      DatasetNotFoundError if data is not found.
    """
    registers = self._get_registers(name)

    # Typically there's only 1, so add special case so that more informative
    # exceptions are raised.
    if len(registers) == 1:
      return registers[0].builder_cls(name)

    # If this dataset has multiple registers, use the first that can be found.
    for register in registers:
      try:
        return register.builder_cls(name)
      except registered.DatasetNotFoundError:
        pass

    raise registered.DatasetNotFoundError(
        f'Namespace {name.namespace} found, '
        f'but could not load dataset {name.name}.'
        f'{self._get_list_builders_context(name)}'
    )

  def builder(
      self,
      name: naming.DatasetName,
      **builder_kwargs: Any,
  ) -> dataset_builder.DatasetBuilder:
    """Loads the builder class for the given dataset."""
    registers = self._get_registers(name)

    # Typically there's only 1, so add special case so that more informative
    # exceptions are raised.
    if len(registers) == 1:
      return registers[0].builder(name, **builder_kwargs)

    if len(registers) > 1:
      raise ValueError(
          f'Namespace {name.namespace} has multiple registers! '
          f'This should not happen! Registers: {registers}'
      )

    raise registered.DatasetNotFoundError(
        f'Namespace {name.namespace} found with {len(registers)} registers, '
        f'but could not load dataset {name.name}.'
    )

  def get_builder_root_dirs(self, name: naming.DatasetName) -> List[epath.Path]:
    """Returns root dir of the generated builder (without version/config)."""
    result = []
    registers = self.registers_per_namespace[name.namespace]
    for register in registers:
      if isinstance(register, register_path.DataDirRegister):
        result.extend(register.get_builder_root_dirs(name))
      else:
        raise RuntimeError(f'Not supported for non datadir registers ({name})!')
    return result


def registry_for_config(config_path: epath.PathLike) -> DatasetRegistry:
  return DatasetRegistry(NamespaceConfig(config_path=epath.Path(config_path)))


community_register = registry_for_config(
    config_path=(utils.tfds_path() / 'community-datasets.toml')
)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Location-based register."""

import concurrent.futures
import difflib
import functools
import os
from typing import Any, Dict, FrozenSet, Iterable, Iterator, List, Type

from absl import flags
from absl import logging
from etils import epath
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import read_only_builder
from tensorflow_datasets.core import registered
from tensorflow_datasets.core.community import register_base
from tensorflow_datasets.core.utils import file_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

TFDS_DEBUG_VERBOSE = flags.DEFINE_boolean(
    'tfds_debug_list_dir', False, 'Debug the catalog generation'
)

# pylint: disable=logging-fstring-interpolation


class DataDirRegister(register_base.BaseRegister):
  """Dataset register based on generated `data_dir` paths.

  Usage:

  ```python
  register = DataDirRegister(namespace_to_data_dirs=
    {'my_namespace': [epath.Path('/path/to/namespaces.toml')]})

  # List all registered datasets: ['kaggle:ds0', 'kaggle:ds1',...]
  register.list_builders()

  # Load a specific dataset
  builder = register.builder('tensorflow_graphics:shapenet')
  ```
  """

  def __init__(self, namespace_to_data_dirs: Dict[str, List[epath.Path]]):
    """Contructor.

    Args:
      namespace_to_data_dirs: Mapping from namespace to list of paths where the
        datasets for that namespace are located.
    """
    self._ns2data_dir = namespace_to_data_dirs

  @functools.cached_property
  def namespaces(self) -> FrozenSet[str]:
    """Available namespaces."""
    return frozenset(self._ns2data_dir)

  def list_builders(self) -> List[str]:
    """Returns the list of registered builders."""
    return sorted(_iter_builder_names(self._ns2data_dir))

  def list_dataset_references(self) -> Iterable[naming.DatasetReference]:
    for namespace, data_dirs in self._ns2data_dir.items():
      for data_dir in data_dirs:
        yield from file_utils.list_datasets_in_data_dir(
            data_dir=data_dir,
            namespace=namespace,
            include_configs=False,
            include_versions=False,
        )

  def builder_cls(
      self,
      name: naming.DatasetName,
  ) -> Type[dataset_builder.DatasetBuilder]:
    """Returns the builder classes."""
    if name.namespace not in self.namespaces:  # pylint: disable=unsupported-membership-test
      error_msg = f'\nNamespace {name.namespace} not found.'
      error_msg += (
          f'Note that namespace should be one of: {sorted(self.namespaces)}'
      )
      raise registered.DatasetNotFoundError(error_msg)
    raise NotImplementedError(
        'builder_cls does not support data_dir-based community datasets. Got: '
        f'{name}'
    )

  def builder(
      self,
      name: naming.DatasetName,
      **builder_kwargs: Any,
  ) -> dataset_builder.DatasetBuilder:
    """Returns the dataset builder."""
    data_dir = builder_kwargs.pop('data_dir', None)
    if data_dir:
      raise ValueError(
          '`data_dir` cannot be set for data_dir-based community datasets. '
          f'Dataset should already be generated. Got: {data_dir}'
      )
    if name.namespace is None:
      raise AssertionError(f'No namespace found: {name}')
    if name.namespace not in self._ns2data_dir:  # pylint: disable=unsupported-membership-test
      close_matches = difflib.get_close_matches(
          name.namespace, self._ns2data_dir, n=1
      )
      hint = f'\nDid you mean: {close_matches[0]}' if close_matches else ''
      error_msg = (
          f'Namespace `{name.namespace}` for `{name}` not found. '
          f'Should be one of {sorted(self._ns2data_dir)}{hint}'
      )
      raise KeyError(error_msg)
    return read_only_builder.builder_from_files(
        name.name,
        data_dir=[
            os.fspath(path) for path in self._ns2data_dir[name.namespace]
        ],
        **builder_kwargs,
    )

  def get_builder_root_dirs(self, name: naming.DatasetName) -> List[epath.Path]:
    """Returns root dir of the generated builder (without version/config)."""
    return [d / name.name for d in self._ns2data_dir[name.namespace]]


def _maybe_iterdir(path: epath.Path) -> Iterator[epath.Path]:
  """Same as `path.iterdir()`, but don't fail if path does not exist."""
  # Use try/except rather than `.exists()` to avoid an extra RPC call
  # per namespace
  try:
    for f in path.iterdir():
      yield f
  except (
      OSError,
      FileNotFoundError,
      PermissionError,
      tf.errors.NotFoundError,
      tf.errors.PermissionDeniedError,
  ) as e:
    pass


def _iter_builder_names(
    ns2data_dir: Dict[str, List[epath.Path]],
) -> Iterator[str]:
  """Yields the `ns:name` dataset names."""
  FILTERED_DIRNAME = frozenset(('downloads',))  # pylint: disable=invalid-name

  def _is_valid_dataset_name(dataset_name: str) -> bool:
    return (
        dataset_name not in FILTERED_DIRNAME
        and naming.is_valid_dataset_name(dataset_name)
    )

  # For better performance, load all namespaces asynchronously
  def _get_builder_names_single_namespace(
      ns_name: str,
      data_dir: epath.Path,
  ) -> List[str]:
    # Note: `data_dir` might contain non-dataset folders, but checking
    # individual dataset would have significant performance drop, so
    # this is an acceptable trade-of.
    return [
        str(naming.DatasetName(namespace=ns_name, name=builder_dir.name))
        for builder_dir in _maybe_iterdir(data_dir)
        if _is_valid_dataset_name(builder_dir.name)
    ]

  with concurrent.futures.ThreadPoolExecutor(max_workers=10) as ex:
    builder_names_futures = []
    for ns_name, data_dirs in ns2data_dir.items():
      for data_dir in data_dirs:
        future = ex.submit(
            _get_builder_names_single_namespace,
            ns_name,
            data_dir,
        )
        builder_names_futures.append(future)

    for future in concurrent.futures.as_completed(builder_names_futures):
      yield from future.result()
